/**************************************************************

CmpackAllFile.c (C-Munipack project)
Readall file input/output
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <float.h>
#include <expat.h>

#include "comfun.h"
#include "console.h"
#include "hash.h"
#include "cmpack_common.h"
#include "cmpack_allfile.h"
#include "allfile.h"
#include "allframes.h"
#include "allparser.h"

/******************* Constants and definitions **********************/

#define BUFFSIZE 	8192
#define ALLOC_BY	64

/********************** Local functions ****************************/

static void file_clean(void *ctx);

/* Alocates memory for photometry file and clears all fields. */
static CmpackAllFile *file_create(void)
{
	CmpackAllFile *f = gc_create(sizeof(CmpackAllFile));
	f->gc.clean = &file_clean;
	all_tab_init(&f->data);
	return f;
}

/* Alocates memory for photometry file and clears all fields. */
static void file_clear(CmpackAllFile *file)
{
	file->aperture = 0;
	file->helcor = 0;
	cmpack_free(file->filter);
	file->filter = NULL;
	all_tab_clear(&file->data);
}

/* Frees all allocated memory in internal structures. */
static void file_clean(void *ctx)
{
	file_clear((CmpackAllFile*)ctx);
	gc_clean(ctx);
}

/* Writes content of the file to the stream in XML format */
static void file_save(CmpackAllFile *f, FILE *to)
{
	int ok;

	/* Header */
    fprintf(to, "# JD, instrumental mags and standard deviations of all detected stars\n");
	fprintf(to, "# ");
	ok = 0;
	if (f->aperture>0) {
		fprintf(to, "Aperture: %d", f->aperture);
		ok = 1;
	}
	if (f->filter) {
		if (ok)
			fprintf(to, ", ");
		fprintf(to, "Filter: %s", f->filter);
		ok = 1;
	}
	if (f->filter) {
		if (ok)
			fprintf(to, ", ");
		fprintf(to, "JD: %s", (f->helcor ? "heliocentric" : "geocentric"));
		ok = 1;
	}
	if (f->data.stars>0) {
		if (ok)
			fprintf(to, ", ");
		fprintf(to, "Stars: %d", f->data.stars);
		ok = 1;
	}
	fprintf(to, "\n");
	/* Data */
	all_tab_write(&f->data, to);
}

/* Writes content of the file to the stream in XML format */
static void file_copy(CmpackAllFile *dst, const CmpackAllFile *src)
{
	dst->aperture = src->aperture;
	cmpack_free(dst->filter);
	dst->filter = (src->filter ? cmpack_strdup(src->filter) : NULL);
	dst->helcor = src->helcor;
	all_tab_copy(&dst->data, &src->data);
}

/* Loads a XML document. Returns zero on success or one of MPK_xxx error codes.
 * Uses Expat library. Parsing context is held in internal file structure,
 * passed as user data (xml_xxx fields).
 */
static int file_load(CmpackAllFile *f, FILE *from)
{
	size_t len;
	int done, res;
	char Buff[BUFFSIZE];
	CmpackParser p;

	all_parser_init(&p, f);
	do {
		len = fread(Buff, 1, BUFFSIZE, from);
		if (ferror(from)) 
			return CMPACK_ERR_READ_ERROR;
		done = feof(from);
		if (all_parser_parse(&p, Buff, (int)len, done)!=0)
   			return CMPACK_ERR_READ_ERROR;
	} while (!done);
	res = p.result;
	all_parser_clear(&p);

	return res;
}

/* Test if the file is a photometry file */
int cmpack_all_test(const char *filename)
{
	int bytes, filesize;
	char buffer[256];

	if (filename) {
		FILE *f = fopen(filename, "rb");
		if (f) {
			fseek(f, 0, SEEK_END);
			filesize = ftell(f);
			fseek(f, 0, SEEK_SET);
			bytes = (int)fread(buffer, 1, 256, f);
			fclose(f);
			return cmpack_all_test_buffer(buffer, bytes, filesize);
		}
	}
	return 0;
}

/* Test if the file is a photometry file */
int cmpack_all_test_buffer(const char *buffer, int buflen, int filesize)
{
	int res;
	char *aux;

	res = 0;
	if (buffer && buflen>32 && filesize>32) {
		/* Convert buffer to a null-terminated string */
		aux = (char*)cmpack_malloc((buflen+1)*sizeof(char));
		memcpy(aux, buffer, buflen);
		aux[buflen] = '\0';
		if (strstr(aux, "# JD, instrumental mags and standard deviations of all detected stars")==aux) {
			/* readall file */
			res = 1;
		}
		cmpack_free(aux);
	}
	return res;
}

/*****************   Public functions    *******************/

/* Create a read-all file context */
CmpackAllFile *cmpack_all_init(void)
{
	return file_create();
}

/* Load data from file. You can open the file for reading only or in 
 * read-write mode. Returns pointer to internal file structure 
 * handle or NULL if failed.
 */
int cmpack_all_load(CmpackAllFile **pfile, const char *filename, CmpackLoadMode flags)
{
	int res;
	FILE *f;
	CmpackAllFile *file;
	
	*pfile = NULL;
	f = fopen(filename, "rb");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;
	file = file_create();
	res = file_load(file, f);
	if (res!=0)
		cmpack_unref(file);
	else
		*pfile = file;
	fclose(f);
	return res;
}

/* Save data to file. You can open the file for reading only or in 
 * read-write mode. Returns pointer to internal file structure 
 * handle or NULL if failed.
 */
int cmpack_all_save(CmpackAllFile *file, const char *filename)
{
	FILE *f = fopen(filename, "wb+");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;
	file_save(file, f);
	fclose(f);
	return 0;
}

/* Copy a table */
int cmpack_all_copy(CmpackAllFile *dstfile, const CmpackAllFile *srcfile)
{
	if (!srcfile || !dstfile) 
		return CMPACK_ERR_INVALID_PAR;

	file_copy(dstfile, srcfile);
	return 0;
}

/* Clear data */
void cmpack_all_clear(CmpackAllFile *file)
{
	file_clear(file);
}


/*********************   Table header     ***********************/

void cmpack_all_set_aperture(CmpackAllFile *file, int aperture)
{
	if (file)
		file->aperture = aperture;
}

int cmpack_all_get_aperture(CmpackAllFile *file)
{
	if (file)
		return file->aperture;
	return 0;
}

void cmpack_all_set_filter(CmpackAllFile *file, const char *filter)
{
	if (file) {
		cmpack_free(file->filter);
		file->filter = (filter ? cmpack_strdup(filter) : NULL);
	}
}

const char *cmpack_all_get_filter(CmpackAllFile *file)
{
	if (file) 
		return file->filter;
	return NULL;
}

void cmpack_all_set_helcor(CmpackAllFile *file, int helcor)
{
	if (file) 
		file->helcor = helcor;
}

int cmpack_all_get_helcor(CmpackAllFile *file)
{
	if (file) 
		return file->helcor;
	return 0;
}

/********************   Frame manipulation    ***********************/

/* Returns number of apertures defined */
int cmpack_all_nframes(CmpackAllFile *f)
{
	if (f)
		return f->data.count;
	return 0;
}

/* Returns number of apertures defined */
int cmpack_all_nstars(CmpackAllFile *f)
{
	if (f)
		return f->data.stars;
	return 0;
}

/* Add new frame */
int cmpack_all_add_frame(CmpackAllFile *file, double jd)
{
	if (file)
		return all_tab_add_frame(&file->data, jd);
	return -1;
}

/* Add new frame */
double cmpack_all_get_jd(CmpackAllFile *file, int index)
{
	if (file) {
		if (index>=0 && index<file->data.count)
			return file->data.list[index].jd;
	}
	return 0;
}

/* Add measurement */
void cmpack_all_set(CmpackAllFile *file, int frame, int star, 
				   int mask, double mag, double err)
{
	if (file) 
		all_tab_set_star(&file->data, frame, star, mask, mag, err);
}

/* Get measurement from the file */
int cmpack_all_get(CmpackAllFile *file, int frame, int star,
		int *mask, double *mag, double *err)
{
	if (mask)
		*mask = 0;

	if (!file) 
		return CMPACK_ERR_INVALID_PAR;
	else if (all_tab_get_star(&file->data, frame, star, mask, mag, err)) 
		return CMPACK_ERR_OK;
	else 
		return CMPACK_ERR_ROW_NOT_FOUND;
}

int cmpack_all_lightcurve(CmpackAllFile *file, int var, int comp, 
		CmpackTable *tab)
{
	int frame, row, vmask, cmask;
	double vmag, cmag, verr, cerr;

	cmpack_tab_clear(tab);
	cmpack_tab_set_type(tab, CMPACK_TABLE_LIGHTCURVE);

	if (var<0 || comp<0 || var>=file->data.stars || comp>=file->data.stars) 
		return CMPACK_ERR_INVALID_PAR;

    /* Table columns */
	cmpack_tab_add_column(tab, "JD", CMPACK_COLUMN_DOUBLE, JD_PRECISION);
	cmpack_tab_add_column(tab, "V-C", CMPACK_COLUMN_DOUBLE, MAG_PRECISION);
	cmpack_tab_add_column(tab, "s1", CMPACK_COLUMN_DOUBLE, MAG_PRECISION);

	/* Table data */
	for (frame=0; frame<file->data.count; frame++) {
		row = cmpack_tab_add_row(tab);
		cmpack_tab_ptdd(tab, row, 0, file->data.list[frame].jd);
		all_tab_get_star(&file->data, frame, var, &vmask, &vmag, &verr);
		all_tab_get_star(&file->data, frame, comp, &cmask, &cmag, &cerr);
		if ((vmask & CMPACK_MASK_MAG) && (cmask & CMPACK_MASK_MAG)) {
			cmpack_tab_ptdd(tab, row, 1, vmag - cmag);
			if ((vmask & CMPACK_MASK_ERR) && (cmask & CMPACK_MASK_ERR)) 
				cmpack_tab_ptdd(tab, row, 2, sqrt(verr*verr + cerr*cerr));
		}
	}
	return 0;
}
