/**************************************************************

amass.c (C-Munipack project)
Air-mass coefficient computation
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_amass.h"

/**********************   PRIVATE DATA TYPES   ***************************/

/* Context for air-mass coefficient computation */
struct _CmpackAirMass
{
	CmpackContext gc;		/**< Global context */
	CmpackConsole *con;		/**< Output console */
	int mask;				/**< Which fields are valid? */
	double ra;				/**< Object's right ascension in hours */
	double dec;				/**< Object's declination in degrees */
	double lat;				/**< Observer's latitude in degrees */
	double lon;				/**< Observer's longitude in degrees */
};

/**********************   PRIVATE FUNCTIONS   ***************************/

static void amass_clean(void *ctx)
{
	CmpackAirMass *lc = (CmpackAirMass*) ctx;

	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/**********************   PUBLIC FUNCTIONS   ***************************/

/* Create local context */
CmpackAirMass *cmpack_amass_init(void)
{
	CmpackAirMass *lc = gc_create(sizeof(CmpackAirMass));
	lc->gc.clean = amass_clean;
	return lc;
}

/* Attach console */
void cmpack_amass_set_console(CmpackAirMass *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set observer's longitude */
void cmpack_amass_set_lon(CmpackAirMass *ctx, double lon)
{
	ctx->mask |= CMPACK_MASK_LON;
	ctx->lon = lon;
}

/* Get observer's longitude */
int cmpack_amass_get_lon(CmpackAirMass *ctx, double *plon)
{
	if (ctx->mask & CMPACK_MASK_LON) {
		if (plon) *plon = ctx->lon;
		return 0;
	} else {
		if (plon) *plon = 0.0;
		return CMPACK_ERR_UNDEF_VALUE;
	}
}

/* Set observer's latitude */
void cmpack_amass_set_lat(CmpackAirMass *ctx, double lat)
{
	ctx->mask |= CMPACK_MASK_LAT;
	ctx->lat = lat;
}

/* Get observer's latitude */
int cmpack_amass_get_lat(CmpackAirMass *ctx, double *plat)
{
	if (ctx->mask & CMPACK_MASK_LAT) {
		if (plat) *plat = ctx->lat;
		return 0;
	} else {
		if (plat) *plat = 0.0;
		return CMPACK_ERR_UNDEF_VALUE;
	}
}

/* Set object's right ascension */
void cmpack_amass_set_ra(CmpackAirMass *ctx, double ra)
{
	ctx->mask |= CMPACK_MASK_RA;
	ctx->ra = ra;
}

/* Get object's right ascension */
int cmpack_amass_get_ra(CmpackAirMass *ctx, double *pra)
{
	if (ctx->mask & CMPACK_MASK_RA) {
		if (pra) *pra = ctx->ra;
		return 0;
	} else {
		if (pra) *pra = 0.0;
		return CMPACK_ERR_UNDEF_VALUE;
	}
}

/* Set object's declination */
void cmpack_amass_set_dec(CmpackAirMass *ctx, double dec)
{
	ctx->mask |= CMPACK_MASK_DEC;
	ctx->dec = dec;
}

/* Get object's declination */
int cmpack_amass_get_dec(CmpackAirMass *ctx, double *pdec)
{
	if (ctx->mask & CMPACK_MASK_DEC) {
		if (pdec) *pdec = ctx->dec;
		return 0;
	} else {
		if (pdec) *pdec = 0.0;
		return CMPACK_ERR_UNDEF_VALUE;
	}
}

/* Append air-mass coefficient to the specified file */
int cmpack_amass(CmpackAirMass *lc, CmpackTable *tab, int jd_column, CmpackAMassFlags flags)
{
	int		count, i, no_airmass, add_altitude, x_column, a_column;
	char	msg[MAXLINE];
	double	jd, x, alt;

	no_airmass		= (flags & CMPACK_AMASS_NOAIRMASS)!=0;
	add_altitude	= (flags & CMPACK_AMASS_ALTITUDE)!=0;

	/* Check parameters */
	if ((lc->mask & CMPACK_MASK_LON)==0) {
		printout(lc->con, 0, "Missing observer's longitude.");
		return CMPACK_ERR_NO_OBS_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_LAT)==0) {
		printout(lc->con, 0, "Missing observer's latitude.");
		return CMPACK_ERR_NO_OBS_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_RA)==0) {
		printout(lc->con, 0, "Missing object's right ascension.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_DEC)==0) {
		printout(lc->con, 0, "Missing object's declination.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	/* Print parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Configuration parameters:");
		sprintf(msg, "\tR.A. = %.3f", lc->ra);
	  	printout(lc->con, 1, msg);
		sprintf(msg, "\tDEC. = %.3f", lc->dec);
	  	printout(lc->con, 1, msg);
		sprintf(msg, "\tLON. = %.3f", lc->lon);
	  	printout(lc->con, 1, msg);
		sprintf(msg, "\tLAT. = %.3f", lc->lat);
	  	printout(lc->con, 1, msg);
  	}

	x_column = -1;
	if (!no_airmass) {
		x_column = cmpack_tab_find_column(tab, "AIRMASS");
		if (x_column<0)
			x_column = cmpack_tab_add_column(tab, "AIRMASS", CMPACK_COLUMN_DOUBLE, AMASS_PRECISION);
	}

	a_column = -1;
	if (add_altitude) {
		a_column = cmpack_tab_find_column(tab, "ALTITUDE");
		if (a_column<0)
			a_column = cmpack_tab_add_column(tab, "ALTITUDE", CMPACK_COLUMN_DOUBLE, ALT_PRECISION);
	}

	count = cmpack_tab_nrows(tab);
	for (i=0; i<count; i++) {
		cmpack_tab_gtdd(tab, i, jd_column, &jd);
		if (jd<=0)
			return CMPACK_ERR_INVALID_JULDAT;

		x = cmpack_airmass(jd, lc->ra, lc->dec, lc->lon, lc->lat, &alt);
		if (is_debug(lc->con)) {
			sprintf(msg, "%.5f -> Alt. = %.3f, X = %.3f", jd, alt, x); 
			printout(lc->con, 1, msg);
		}
		
		if (x_column)
			cmpack_tab_ptdd(tab, i, x_column, x);
		if (a_column)
			cmpack_tab_ptdd(tab, i, a_column, alt);
	}

	/* Normal return */
	return 0;
}

/* Compute air-mass coefficient for single julian date */
int cmpack_amass_jd(CmpackAirMass *lc, double jd, double *altitude, double *amass)
{
	char   msg[MAXLINE], buf[MAXLINE];
	double x, alt;

	/* Check parameters */
	if ((lc->mask & CMPACK_MASK_LON)==0) {
		printout(lc->con, 0, "Missing observer's longitude.");
		return CMPACK_ERR_NO_OBS_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_LAT)==0) {
		printout(lc->con, 0, "Missing observer's latitude.");
		return CMPACK_ERR_NO_OBS_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_RA)==0) {
		printout(lc->con, 0, "Missing object's right ascension.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_DEC)==0) {
		printout(lc->con, 0, "Missing object's declination.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}

	/* Print parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Configuration parameters:");
	  	cmpack_ratostr(lc->ra, buf, MAXLINE);
		printpars(lc->con, "R.A.", 1, buf);
		cmpack_dectostr(lc->dec, buf, MAXLINE);
		printpars(lc->con, "DEC.", 1, buf);
	  	cmpack_lontostr(lc->lon, buf, MAXLINE);
		printpars(lc->con, "LON.", 1, buf);
	  	cmpack_lattostr(lc->lat, buf, MAXLINE);
		printpars(lc->con, "LAT.", 1, buf);
  	}
	if (jd>=2400000 && jd<3000000) {
		/* Full julian date */
		x = cmpack_airmass(jd, lc->ra, lc->dec, lc->lon, lc->lat, &alt);
		if (is_debug(lc->con)) {
			sprintf(msg, "%.5f -> Alt. = %.3f, X = %.3f", jd, alt, x); 
			printout(lc->con, 1, msg);
		}
		if (amass)
			*amass = x;
		if (altitude)
			*altitude = alt;
		return 0;
	}
	return CMPACK_ERR_INVALID_JULDAT;
}
