/**************************************************************

cfgfile.c (C-Munipack project)
Copyright (C) 2004 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <expat.h>

#include "cmpack_common.h"
#include "cmpack_catfile.h"
#include "catparser.h"
#include "catfile.h"

#define BUFFSIZE 4096

/* Context structure for catalogue file parser */
struct _CmpackCatParser
{
	CmpackCatFile *file;				/**< Catalogue file context */
	int head_only;						/**< Load header only */
	int state;							/**< Parser state */
	int result;							/**< Parser result code */
	int level;							/**< Parser level */
	char *element;						/**< Current element */
};
typedef struct _CmpackCatParser CmpackCatParser;

/**********************   HELPER FUNCTIONS   *****************************/

/* Returns pointer to the buffer which holds the value of the XML attribute.
 * The 'key' parameter specifies the name of the attribute. Do not free the
 * memory.
*/
static const char *get_attr(const char **atlist, const char *key)
{
	int i;
	
  	for (i=0; atlist[i]; i+=2) {
	  	if (strcmp(atlist[i], key)==0)
		  	return atlist[i+1];
  	}
  	return NULL;
}	

/*********************   XML PARSER HANDLERS   *************************/

/* Obligatory callback from XML parser */
static int XMLCALL XMLUnknownEncodingHandler(void *data,const XML_Char *encoding,XML_Encoding *info)
{
 	int i;
 	
 	for (i=0; i<256; i++) 
 		info->map[i] = i;
 	info->data = NULL;
 	info->convert = NULL;
 	info->release = NULL;
 	return XML_STATUS_OK;
}

/* Start element callback. This is called whenever the XML parser finds
 * a start of the element. The 'aname' parameter is name of the element,
 * 'atlist' is the list of attributes (key+val) pairs.
*/
static void XMLCALL XMLStartElementHandler(void *adata, const char *aname, const char **atlist)
{
	int		xok, yok, mask, id;
	double	x, y, mag;
	char	*aux, *endptr, *aux1, *aux2;
  	CmpackCatParser *f;
	CmpackCatStar *st;

	f = (CmpackCatParser*) adata;
	switch (f->state) 
	{
	case 0:
		/* Root element */
		if (strcmp(aname, "cat_file")==0) {
			f->state = 1;
			f->level = 0;
			break;
		}
		f->result = CMPACK_ERR_UNKNOWN_FORMAT;
		break;

	case 1:
		/* First level elements */
		if (strcmp(aname, "info")==0) {
			f->level++;
			f->state = 2;
			break;
		}
		if (strcmp(aname, "selection")==0) {
			/* Apertures */
			f->level++;
			f->state = 4;
			break;
		}
		if (strcmp(aname, "stars")==0 && !f->head_only) {
			/* Stars */
			cat_stars_set_size(f->file->stars, 
				atoi(get_attr(atlist, "width")),
				atoi(get_attr(atlist, "height")));
			f->level++;
			f->state = 5;
			break;
		}
		/* Unknown section -> ignore it */
		if (f->state==1) {
			f->level++;
			f->state = 3;
			break;
		}
		break;
		
	case 2:
		/* Read header <key>value</key> */
		if (f->level==1) {
			f->element = cmpack_realloc(f->element, strlen(aname)+1);
			strcpy(f->element, aname);
		}
		f->level++;
		break;

	case 3:
		/* Ignore unknown header */
		f->level++;
		break;

	case 4:
		/* Selection: <s id="n" label="xxx"/> */
		if (f->level==1 && strcmp(aname, "select")==0) {
			aux1 = (char*) get_attr(atlist, "id");
			aux2 = (char*) get_attr(atlist, "label");
			if (aux1 && aux2) {
				if (strstr(aux2, "var")==aux2)
					cat_seltab_set_star(f->file->select, atoi(aux1), CMPACK_SELECT_VAR);
				else if (strstr(aux2, "comp")==aux2)
					cat_seltab_set_star(f->file->select, atoi(aux1), CMPACK_SELECT_COMP);
				else if (strstr(aux2, "chk")==aux2)
					cat_seltab_set_star(f->file->select, atoi(aux1), CMPACK_SELECT_CHECK);
			}
		}
		f->level++;
		break;
		
	case 5:
		/* Stars: <s x="val1" y="val2" m="mag" e="err"/> */
		if (f->level==1 && strcmp(aname, "s")==0) {
			mask = 0;
			aux = (char*) get_attr(atlist, "id");
			if (aux) {
				id = strtol(aux, &endptr, 10);
				if (aux!=endptr) {
					x = y = -1.0;
					mag = 99.999;
					xok = yok = 0;
					aux = (char*) get_attr(atlist, "x");
					if (aux) {
						x = strtod(aux, &endptr);
						xok = (aux!=endptr);
					}
					aux = (char*) get_attr(atlist, "y");
					if (aux) {
						y = strtod(aux, &endptr);
						yok = (aux!=endptr);
					}
					if (xok && yok) 
						mask |= CMPACK_MASK_POS;
					aux = (char*) get_attr(atlist, "m");
					if (aux) {
						mag = strtod(aux, &endptr);
						if (aux!=endptr) 
							mask |= CMPACK_MASK_MAG;
					}
					if (id>=0 && (mask & CMPACK_MASK_POS)) {
						st = cat_star_new(id);
						if (st) {
							cat_star_set(st, mask, x, y, mag);
							cat_stars_add(f->file->stars, st);
						}
					}
				}
			}
		}
		f->level++;
		break;
		
	default:
		f->result = CMPACK_ERR_READ_ERROR;
	}
}

/* Character data handler */
static void XMLCALL XMLCharacterDataHandler(void *adata, const char *s, int len)
{
	int i;
  	CmpackCatParser *f;
  	
	f = (CmpackCatParser*) adata;
	if (f->state==2 && f->level==2 && f->element) {
		char *str = cmpack_malloc(len+1);
		if (str) {
			memcpy(str, s, len);
			/* Trim trailing spaces */
			while (len>0 && str[len-1]==' ')
				len--;
			str[len] = '\0';
			/* Trim leading spaces */
			i = (int)strspn(str, " ");
			if (i>0) 
				memmove(str, str+i, len-i+1);
			cat_head_set(f->file->head, f->element, str, NULL);
			cmpack_free(str);
		}
	}
}
 

/* End element handler. XML parser calls this function whenever the end of
 * an element is found.
*/
static void XMLCALL XMLEndElementHandler(void *adata, const char *aname)
{
  	CmpackCatParser *f;
		
	f = (CmpackCatParser*) adata;
	switch (f->state) 
	{
	case 2:
	case 3:
	case 4:
	case 5:
		f->level--;
		if (f->level == 0)
			f->state = 1;
		break;

	case 102:
	case 112:
	case 122:
	case 103:
	case 104:
	case 105:
		f->level--;
		if (f->level == 0)
			f->state = 101;
		break;
		
	case 106:
		f->level--;
		if (f->level == 1)
			f->state = 105;
		break;
	}
}

/********************   PUBLIC FUNCTIONS   ***************************/

/* Loads a XML document. Returns zero on success or one of MPK_xxx error codes.
 * Uses Expat library. Parsing context is held in internal file structure,
 * passed as user data (xml_xxx fields).
 */
int cat_loadxml(CmpackCatFile *f, FILE *from, CmpackLoadMode flags)
{
	int done, len;
	char buf[BUFFSIZE];
	CmpackCatParser p;
	XML_Parser parser = XML_ParserCreate(NULL);

	memset(&p, 0, sizeof(CmpackCatParser));
	p.file = f;
	p.head_only = flags & CMPACK_LOAD_HEADONLY;
	XML_SetUserData(parser, &p);
	XML_SetElementHandler(parser, XMLStartElementHandler, XMLEndElementHandler);
	XML_SetCharacterDataHandler(parser, XMLCharacterDataHandler);
	XML_SetUnknownEncodingHandler(parser, XMLUnknownEncodingHandler, NULL);
	do {
		len = (int)fread(buf, 1, BUFFSIZE, from);
		if (ferror(from)) 
			return CMPACK_ERR_READ_ERROR;
		done = feof(from);
		if (XML_Parse(parser, buf, (int)len, done) == XML_STATUS_ERROR)
   			return CMPACK_ERR_READ_ERROR;
	} while (!done);
	cmpack_free(p.element);
	XML_ParserFree(parser);
	return p.result;
}
