/**************************************************************

iost.c (C-Munipack project)
SBIG files low-level routines
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "comfun.h"
#include "cmpack_common.h"
#include "iost.h"

/* header length of the SBIG image file */
#define STHEADSIZE   2048             

/* File descriptor */
struct stfile
{
	FILE *f;
	char *head;
	int width, height;
	unsigned short *data;
};

/* Compressed or uncompressed file? */
static int stgetftype(const char *head)
{
	char *aux, *aux2;
	
	if ((strstr(head,"ST-")==head)||(strstr(head,"SBIG")==head)||(strstr(head,"PixCel")==head)) {
		aux  = strchr(head,0x0A);
		aux2 = strstr(head,"Compressed");
		if (aux!=NULL) {
			/* SBIG compressed or uncompressed file */
			return ((aux2!=NULL)&&(aux2<aux)?2:1);
		}
	}
	/* This is not SBIG file */
	return -1;
}

/* Seek routine both for file and memory buffer */
static int stseek(struct stfile *st, long offset, int base)
{
	return fseek(st->f, offset, base);
}

/* Seek routine both for file and memory buffer */
static size_t stread(void *buffer, size_t size, size_t num, struct stfile *st)
{
	return fread(buffer, size, num, st->f);
}

/* Open the file and read the header into memory */
int stopen(struct stfile **phandle, const char *filename)
{
	size_t headlen;
	struct stfile *st;

	*phandle = NULL;

	/* open the file */
	st = (struct stfile*) cmpack_calloc(1, sizeof(struct stfile));
	st->f = fopen(filename,"rb");
	if (st->f==NULL) {
		/* File not found */
		cmpack_free(st);
		return CMPACK_ERR_OPEN_ERROR;
	}

	/* Read header of the file */
	st->head = (char *)cmpack_malloc(STHEADSIZE);
	headlen = fread(st->head,1,STHEADSIZE,st->f);

	/* Identify the file type */
	if ((headlen!=STHEADSIZE)||(stgetftype(st->head)<0)) {
		/* Invalid file header */
		fclose(st->f);
		cmpack_free(st->head);
		cmpack_free(st);
		return CMPACK_ERR_READ_ERROR; 
	}

	/* Normal return */
	*phandle = st;
	return 0;
}

/* Close a file and free allocated memory */
void stclos(struct stfile *st)
{
	if (st) {
		if (st->f) 
			fclose(st->f);
		cmpack_free(st->head);
		cmpack_free(st->data);
	}
	cmpack_free(st);
}

/* Get string by key */
int stgkys(struct stfile *st, const char *key, char **buf)
{
	size_t len;
	char *aux, *j;

	if (buf) *buf = NULL;
	
	/* Locate a key in the header */
	aux = (char *)cmpack_malloc(strlen(key)+3);
	strcpy(aux,key);
	strcat(aux," =");
	j = strstr(st->head,aux);
	cmpack_free(aux);
	if (j==NULL) {
		/* The key was not found */
		return CMPACK_ERR_KEY_NOT_FOUND;   
	}
	/* Copy the value to buf */
	if (buf) {
		j = j + strlen(key) + 2;
		len = strcspn(j, "\r\n\x1A");
		if (len>0) {
			*buf = (char*)cmpack_malloc(len+1);
			memcpy(*buf, j, len);
		}
		(*buf)[len] = '\0';
	}
	return 0;
}

/* Get integer number by key */
int stgkyi(struct stfile *st, const char *key, int *value)
{
	char *aux, *j;
	
	*value = 0;
	/* Locate a key in the header */
	aux = (char *)cmpack_malloc(strlen(key)+3);
	strcpy(aux,key);
	strcat(aux," =");
	j = strstr(st->head,aux);
	cmpack_free(aux);
	if (j==NULL) {
		/* The key was not found */
		return CMPACK_ERR_KEY_NOT_FOUND; 
	}
	/* Convert a value */
	j = j + strlen(key) + 2;
	*value = atol(j);
	return 0;
}

/* Get floating number by key */
int stgkyd(struct stfile *st, const char *key, double *value)
{
	char *aux, *j;
	
	*value = 0.0;
	/* Locate a key in the header */
	aux = (char *)cmpack_malloc(strlen(key)+3);
	strcpy(aux,key);
	strcat(aux," =");
	j = strstr(st->head,aux);
	cmpack_free(aux);
	if (j==NULL) {
		/* The key was not found */
		return CMPACK_ERR_KEY_NOT_FOUND; 
	}
	/* Convert a value */
	j = j + strlen(key) + 2;
	*value = atof(j);
	return 0;
}

/* Get key name and value by index */
int stgkyn(struct stfile *st,  int nkey, char **keyname, char **keyval) 
{
	size_t keylen, vallen;
	char *aux, *j, buf[MAXLINE];
	
	if (keyname) *keyname = NULL;
	if (keyval)  *keyval = NULL;
	/* Locate a nth line in the header */
	aux = st->head;
	while ((aux!=NULL)&&(nkey>0)) {
		j = strchr(aux,0x0A);
		if (j) {
			aux = j+2;
			nkey--;
		}
	}
	if (aux==NULL) {
		/* Beyond end of header */
		return CMPACK_ERR_KEY_NOT_FOUND;
	}
	/* Extract the line from the header */
	j = strchr(aux,0x0A);
	strncpy(buf,aux,j-aux);
	buf[j-aux] = '\0';
	if (strcmp(buf,"End")==0) {
		/* Beyond end of header */
		return CMPACK_ERR_KEY_NOT_FOUND;
	}
	/* Parse buf into keyname, keyval (comm is always empty) */
	j = strchr(buf,'=');
	if (j) {
		keylen = (j-buf-1);
		if (keylen>0 && keyname) {
			*keyname = (char*)cmpack_malloc(keylen+1);
			memcpy(*keyname, buf, keylen);
			(*keyname)[keylen] = '\0';
		}
		vallen = strlen(j+2);
		if (vallen>0 && keyval) {
			*keyval = (char*)cmpack_malloc(vallen+1);
			memcpy(*keyval, j+2, vallen);
			(*keyval)[vallen] = '\0';
		}
	} else {
		*keyname = cmpack_strdup(buf);
	}
	return 0;
}

/* Get image size */
int stghpr(struct stfile *st, int *nx, int *ny)
{
	stgkyi(st, "Width", nx);
	stgkyi(st, "Height", ny);
	return 0;
}

/* Decode image data */
static int loadimage(struct stfile *st, unsigned short **pdata, int *pwidth, int *pheight)
{
	int nx, ny;
	int i, j, k, deltai, hodnota;
	size_t headlen, bytes;
	unsigned short delka;
	unsigned char *buf;

	if (st->data) {
		*pdata = st->data;
		stgkyi(st, "Width", pwidth);
		stgkyi(st, "Height", pheight);
		return 1;
	}

	/* Get width and height of the image */
	stgkyi(st, "Width", &nx);
	stgkyi(st, "Height", &ny);
	if ((nx<=0)||(ny<=0)||(nx>=100000)||(ny>=100000)) {
		/* Cannot read the width or height of the image */
		return 0;     
	}
	/* Read image data */
	st->data = (unsigned short *)cmpack_calloc(nx*ny, sizeof(unsigned short));
	stseek(st, STHEADSIZE, SEEK_SET);
	if (stgetftype(st->head)!=2) {
		/* uncompressed image */
		bytes = nx*ny;
		if (stread(st->data,2,bytes,st)!=bytes) {
			/* invalid image block */
			cmpack_free(st->data);
			st->data = NULL;
			return 0;                            
		}
	} else {
		/* Compressed image, courtesy by P. Pravec */
		buf = (unsigned char *)cmpack_malloc(2*nx);
		for (i=0;i<ny;i++) {
			/* for each line */
			/* length of line in bytes */
			headlen = stread(&delka,1,2,st);
			if ((headlen<2)||(delka>2*nx)) {
				/* invalid row length */
				cmpack_free(buf);
				cmpack_free(st->data);
				st->data = NULL;
				return 0;
			} else
			if (delka==2*nx) {
				/* no compressed line in compressed image */
				headlen = stread(&st->data[i*nx],1,2*nx,st);
				if (headlen!=2*nx) {
					/* invalid row length */
					cmpack_free(buf);
					cmpack_free(st->data);
					st->data = NULL;
					return 0;
				}
			} else {
				/* compressed line */
				headlen = stread(buf,1,delka,st);
				if (headlen!=delka) {
					/* invalid row length */
					cmpack_free(buf);
					cmpack_free(st->data);
					st->data = NULL;
					return 0;
				}
				/* first value in the line */
				hodnota = buf[0] + 256*buf[1]; 
				st->data[i*nx] = hodnota;
				j = 1; k = 2;
				while (k<delka) {
					deltai = buf[k];
					k++;
					if (deltai>127) deltai -= 256;
					if (deltai==-128) {
						hodnota=buf[k]+256*buf[k+1];
						k+=2;
					} else {
						hodnota +=deltai;
					}
					st->data[i*nx+j] = hodnota;
					j++;
				}
			}
		}
		cmpack_free(buf);
	}

	/* Normal return */
	*pdata = st->data;
	*pwidth = nx;
	*pheight = ny;
	return 1;
}

int stg2duj(struct stfile *st, int flipx, int flipy, unsigned short *ccd)
{
	unsigned short *data;
	int i, j, nx, ny;
	
	if (loadimage(st, &data, &nx, &ny)) {
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				ccd[j*nx+i] = st->data[j*nx+i];
			}
		}
		return 0;
	}
	return CMPACK_ERR_READ_ERROR;
}

int stg2dl(struct stfile *st, int flipx, int flipy, long *ccd)
{
	unsigned short *data;
	int i, j, nx, ny;
	
	if (loadimage(st, &data, &nx, &ny)) {
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				ccd[j*nx+i] = st->data[j*nx+i];
			}
		}
		return 0;
	}
	return CMPACK_ERR_READ_ERROR;
}

int stg2dul(struct stfile *st, int flipx, int flipy, unsigned long *ccd)
{
	unsigned short *data;
	int i, j, nx, ny;
	
	if (loadimage(st, &data, &nx, &ny)) {
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				ccd[j*nx+i] = st->data[j*nx+i];
			}
		}
		return 0;
	}
	return CMPACK_ERR_READ_ERROR;
}

int stg2de(struct stfile *st, int flipx, int flipy, float *ccd)
{
	unsigned short *data;
	int i, j, nx, ny;
	
	if (loadimage(st, &data, &nx, &ny)) {
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				ccd[j*nx+i] = st->data[j*nx+i];
			}
		}
		return 0;
	}
	return CMPACK_ERR_READ_ERROR;
}

int stg2dd(struct stfile *st, int flipx, int flipy, double *ccd)
{
	unsigned short *data;
	int i, j, nx, ny;
	
	if (loadimage(st, &data, &nx, &ny)) {
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				ccd[j*nx+i] = st->data[j*nx+i];
			}
		}
		return 0;
	}
	return CMPACK_ERR_READ_ERROR;
}

/* Reads observation date */
int stgetdate(struct stfile *st, int *yr, int *mon, int *day)
{
    int y=0, y2k=0, m=0, d=0;
    char *datestr;

	if (stgkys(st, "Date", &datestr)!=0)
		return CMPACK_ERR_INVALID_DATE;
    if (stgkyi(st, "Y2KYear", &y2k)!=0) 
		y2k = 0;

	/* Parse input date and time */
	sscanf(datestr, "%2d%*1s%2d%*1s%d", &m, &d, &y);
	cmpack_free(datestr);

	/* Compute the year */
	*mon = m;
	*day = d;
	if (y2k>0)
		*yr = y2k;
	else if ((y>50)&&(y<=999))
		*yr = y + 1900;
	else if ((y>=0)&&(y<50))
		*yr = y + 2000;
	else if ((y>=1900)&&(y<1950))
		*yr = y + 100;
	else
		*yr = y;

	/* Check values */
	if ((*yr>1900)&&(*yr<=9999)&&(*mon>=1)&&(*mon<=12)&&(*day>=1)&&(*day<=31))
		return 0;
	else
		return CMPACK_ERR_INVALID_DATE;
}

/* Reads observation time */
int stgettime(struct stfile *st, int *hr, int *min, int *sec)
{
    char *timestr;

	if (stgkys(st, "Time", &timestr)!=0)
		return CMPACK_ERR_INVALID_DATE;

	/* Parse input date and time */
	sscanf(timestr,"%2d%*1s%2d%*1s%2d",hr,min,sec);
	cmpack_free(timestr);

	/* Check values */
	if ((*hr>=0)&&(*hr<=23)&&(*min>=0)&&(*min<=59)&&(*sec>=0)&&(*sec<=59))
		return 0;
	else
		return CMPACK_ERR_INVALID_DATE;
}
