/**************************************************************

konvfits.c (C-Munipack project)
Conversion functions for FITS files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <float.h>

#include "comfun.h"
#include "console.h"
#include "trajd.h"
#include "config.h"
#include "cmpack_common.h"
#include "konvfits.h"

struct _CmpackFitsFile
{
	fitsfile *fits;
	int status;
	char *buffer;
	size_t bufsize;
};
typedef struct _CmpackFitsFile CmpackFitsFile;

/* Realloc procedure */
static void *realloc_proc(void *ptr, size_t len)
{
	return cmpack_realloc(ptr, len);
}

/* Reads date */
static int getdatetime(fitsfile *fits, CmpackDateTime *dt, int *end, int *center)
{
	int x, status, date_ok, time_ok;
	char datestr[FLEN_VALUE], timestr[FLEN_VALUE], *ptr;

	memset(dt, 0, sizeof(CmpackDateTime));
	*end = *center = 0;

	status = 0;
	ffgkys(fits,"DATE-OBS",datestr,NULL,&status);
    if (status!=0) { strcpy(datestr,""); status=0; }
    if (strlen(datestr)==0) {
	    /* D02E-OBS= 'dd/mm/yy' */
	    ffgkys(fits,"D02E-OBS",datestr,NULL,&status);
        if (status!=0) { strcpy(datestr,""); status=0; }
    }
    if (strlen(datestr)==0) {
	    /* DATE-BEG= '2005-08-24T20:19:03' */
		ffgkys(fits,"DATE-BEG",datestr,NULL,&status);
    	if (status!=0) { strcpy(datestr,""); status=0; }
    }
    if (strlen(datestr)==0) {
	    /* DATE_OBS= '2005-08-24T20:19:03' */
		ffgkys(fits,"DATE_OBS",datestr,NULL,&status);
    	if (status!=0) { strcpy(datestr,""); status=0; }
    }

	ptr = datestr;
	while (*ptr==' ')
		ptr++;
    
    date_ok = 0;
    if ((strlen(ptr)==8)&&(ptr[2]=='/')&&(ptr[5]=='/')) {
	    /* dd/mm/yy */
		dt->date.year = 2000 + atoi(ptr+6);
		dt->date.month = atoi(ptr+3);
		dt->date.day = atoi(ptr);
		date_ok = 1;
    } else
    if (strlen(ptr)==10 && ptr[2]=='/' && ptr[5]=='/') {
    	/* dd/mm/yyyy */
		dt->date.year = atoi(ptr+6);
		dt->date.month = atoi(ptr+3);
		dt->date.day = atoi(ptr);
		date_ok = 1;
    } else 
	if (strlen(ptr)>=10 && ptr[4]=='-' && ptr[7]=='-') {
		/* yyyy-mm-dd */
        dt->date.year = atoi(ptr);
        dt->date.month = atoi(ptr+5);
        dt->date.day = atoi(ptr+8);
		date_ok = 1;
    } else 
	if (strlen(ptr)>=10 && ptr[4]=='/' && ptr[7]=='/') {
		/* yyyy/mm/dd */
        dt->date.year = atoi(ptr);
        dt->date.month = atoi(ptr+5);
        dt->date.day = atoi(ptr+8);
		date_ok = 1;
	}

	status = 0;
	ffgkys(fits,"TIME-OBS",timestr,NULL,&status);
	if (status!=0) { strcpy(timestr,""); status=0; }
    if (strlen(timestr)==0) {
	    /* DATE-BEG= '2005-08-24T20:19:03' */
		ffgkys(fits,"DATE-BEG",datestr,NULL,&status);
    	if (status!=0) { strcpy(datestr,""); status=0; }
		ptr = strchr(datestr, 'T');
		if (ptr) 
  			strcpy(timestr,ptr+1);
    }
    if (strlen(timestr)==0) {
        /* DATE-OBS= '2004-03-20T22:55:28.014' / GMT START OF EXPOSURE */
		ffgkys(fits,"DATE-OBS",datestr,NULL,&status);
		if (status!=0) { strcpy(datestr,""); status=0; }
		ptr = strchr(datestr, 'T');
		if (ptr) 
  			strcpy(timestr,ptr+1);
    }
    if (strlen(timestr)==0) {
        /* DATE_OBS= ' 2004-03-20T22:55:28.014' / GMT START OF EXPOSURE */
		ffgkys(fits,"DATE_OBS",datestr,NULL,&status);
		if (status!=0) { strcpy(datestr,""); status=0; }
		ptr = strchr(datestr, 'T');
		if (ptr) 
  			strcpy(timestr,ptr+1);
    }
    if (strlen(timestr)==0) {
	    /* TM-START=  1.3079446944444E+01 / [Corrected] UT obs'n start, real hours */
	    ffgkys(fits,"TM-START",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
    }
    if (strlen(timestr)==0) {
		/* REC-STRT= '13:04:46'           / Recorded UT time of observation start */
        ffgkys(fits,"REC-STRT",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
	}
    if (strlen(timestr)==0) {
        ffgkys(fits,"UT",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
    }
    if (strlen(timestr)==0) {
	    /* UT_END  = '19:33:59'				/ PMIS Image Processing Software */
		ffgkys(fits,"UT_END",timestr,NULL,&status);
		if (status!=0) { strcpy(timestr,""); status=0; }
		else *end = 1;
    }
    if (strlen(timestr)==0) {
        ffgkys(fits,"UT-START",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
    }
    
    time_ok = 0;
	if ((strlen(timestr)>=8)&&(timestr[2]==':')&&(timestr[5]==':')) {
		dt->time.hour = atol(timestr);
		dt->time.minute = atol(timestr+3);
        x = (int)(1000.0*atof(timestr+6)+0.5);
		dt->time.second = x/1000;
		dt->time.milisecond = x%1000;
		time_ok = 1;
	}
	if (!time_ok && !strchr(timestr, ':') && !strchr(timestr, ' ') && strchr(timestr, '.')) {
		double val = strtod(timestr, &ptr);
		if (ptr!=timestr) {
			int msec = (int)(val*3600000.0+0.5);
			dt->time.hour = (msec/3600000);
			dt->time.minute = (msec/60000)%60;
			dt->time.second = (msec/1000)%60;
			dt->time.milisecond = msec%1000;
			time_ok = 1;
		}
	}
	
	return (date_ok && time_ok);
}

static int getexptime(fitsfile *fits, double *value)
{
	char buf[FLEN_VALUE];
	int stat = 0;

	*value=0.0;

	ffgkyd(fits, "EXPTIME", value, NULL, &stat);
	if (stat==0) 
		return 1;
	stat = 0;
	ffgkys(fits, "EXPOSURE", buf, NULL, &stat);
	if (stat==0) { 
		*value = atof(buf); 
		return 1; 
	}
	stat = 0;
	ffgkys(fits, "ITIME", buf, NULL, &stat);
	if (stat==0) { 
		*value = atof(buf); 
		return 1; 
	}

	return 0;
}

/* Copy frame header, skip FITS specific keywords */
static int copyheader(fitsfile *src, fitsfile *dst)
{
	int i, status = 0, dstat = 0;
	char key[FLEN_KEYWORD], val[FLEN_VALUE], com[FLEN_COMMENT];
	char card[FLEN_CARD];

	for (i=1; ffgkyn(src, i, key, val, com, &status)==0; i++) {
		if (strcmp(key, "SIMPLE")==0 || strcmp(key, "BITPIX")==0 ||
			strstr(key, "NAXIS")==key || strcmp(key, "EXTEND")==0 ||
			strcmp(key, "BZERO")==0 || strcmp(key, "BSCALE")==0 ||
			strcmp(key, "END")==0 || (strcmp(key, "COMMENT")==0 && (
				strstr(com, "FITS (Flexible Image Transport System) format is defined in 'Astronomy") ||
				strstr(com, "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H")))) {
				/* Skip this keyword */
		} else {
			ffgrec(src, i, card, &status);
			if (status==0)
				ffprec(dst, card, &dstat);
		}
		if (dstat!=0)
			return CMPACK_ERR_WRITE_ERROR;
	}
	return 0;
}

/* Copy frame header, skip FITS specific keywords */
static int clearheader(fitsfile *fits)
{
	int i, status = 0;
	char key[FLEN_KEYWORD], val[FLEN_VALUE], com[FLEN_COMMENT];

	i = 1;
	while (ffgkyn(fits, i, key, val, com, &status)==0) {
		if (strcmp(key, "SIMPLE")==0 || strcmp(key, "BITPIX")==0 ||
			strstr(key, "NAXIS")==key || strcmp(key, "EXTEND")==0 ||
			strcmp(key, "BZERO")==0 || strcmp(key, "BSCALE")==0 ||
			strcmp(key, "END")==0 || (strcmp(key, "COMMENT")==0 && (
				strstr(com, "FITS (Flexible Image Transport System) format is defined in 'Astronomy") ||
				strstr(com, "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H")))) {
				/* Skip this keyword */
					i++;
		} else {
			ffdrec(fits, i, &status);
		}
	}
	return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
}

/* Write file origin */
static void setorigin(fitsfile *fits)
{
	int stat = 0;
	char val[FLEN_VALUE];

	sprintf(val, "%s %s", PACKAGE_NAME, VERSION);
	ffukys(fits, "ORIGIN", val, "CREATOR NAME AND VERSION", &stat);
}

/* Write date of creation */
static void setpcdate(fitsfile *fits)
{
	int stat = 0;
	time_t tsec;
	struct tm *actim;
	char val[FLEN_VALUE];

	tsec = time(&tsec);
	actim = localtime(&tsec);
	sprintf(val,"%04d-%02d-%02d",actim->tm_year+1900,actim->tm_mon+1,actim->tm_mday);
 	ffukys(fits, "DATE", val, "DATE OF CREATION", &stat);
}

/* Returns nonzero if the file is of FITS format */
int fits_test(const char *block, size_t length)
{
  return (length>2880 && strstr(block, "SIMPLE  =")==block);
}

/* Open a FITS file */
int fits_open(tHandle *handle, const char *filename, CmpackOpenMode flags)
{
	int readonly;

	CmpackFitsFile *fs = (CmpackFitsFile*)cmpack_calloc(1, sizeof(CmpackFitsFile));
	if (flags & CMPACK_OPEN_CREATE) {
		remove(filename);
		ffdkinit(&fs->fits, filename, &fs->status);
	} else {
		readonly = (flags & CMPACK_OPEN_READONLY)!=0;
		ffdkopn(&fs->fits, filename, (readonly ? READONLY : READWRITE), &fs->status);
	}
	if (fs->status!=0) {
		cmpack_free(fs);
		*handle = NULL;
		return CMPACK_ERR_OPEN_ERROR;
	} else {
		*handle = fs;
		return 0;
	}
}

/* Open a memory file */
tHandle fits_init(void)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)cmpack_calloc(1, sizeof(CmpackFitsFile));
	ffimem(&fs->fits, (void*)&fs->buffer, &fs->bufsize, 0x4000, &realloc_proc, &fs->status);
	if (fs->status!=0) {
		cmpack_free(fs);
		return NULL;
	}
	return fs;
}

/* Close file and free allocated memory */
void fits_close(tHandle handle)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	if (fs->fits) 
		ffclos(fs->fits, &fs->status);
	if (fs->buffer)
		cmpack_free(fs->buffer);
	cmpack_free(fs);
}

/* Close file and free allocated memory */
fitsfile *fits_getfits(tHandle handle)
{
	return ((CmpackFitsFile*)handle)->fits;
}

/* Reads string value from header */
int fits_gkys(tHandle handle, const char *key, char **buf)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	char val[FLEN_VALUE];
	int status = 0;
	
	ffgkys(fs->fits, (char*) key, val, NULL, &status);
	if (status==0) {
		if (buf) *buf = cmpack_strdup(val);
		return 0;
	} else {
		if (buf) *buf = NULL;
		return CMPACK_ERR_KEY_NOT_FOUND;
	}
}

/* Reads integer value from header */
int fits_gkyi(tHandle handle, const char *key, int *val)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int status = 0;
	long lval;

	ffgkyj(fs->fits, (char*) key, &lval, NULL, &status);
	if (status==0) {
		*val = (int)lval;
		return 0;
	} else {
		*val = 0;
		return CMPACK_ERR_KEY_NOT_FOUND;
	}
}

/* Reads integer value from header */
int fits_gkyl(tHandle handle, const char *key, int *val)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int status = 0;

	ffgkyl(fs->fits, (char*) key, val, NULL, &status);
	if (status==0) {
		return 0;
	} else {
		*val = 0;
		return CMPACK_ERR_KEY_NOT_FOUND;
	}
}

int fits_getsize(tHandle handle, int *width, int *height)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int status = 0;
    long naxes[2];

	ffgisz(fs->fits, 2, naxes, &status);
	if (status==0) {
		if (width) *width = naxes[0];
		if (height) *height = naxes[1];
		return 0;
	} else {
		if (width) *width = 0;
		if (height) *height = 0;
		return CMPACK_ERR_INVALID_SIZE;
	}
}

/* Get image data format */
CmpackBitpix fits_getbitpix(tHandle handle)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int imgtype, status = 0;

	ffgiet(fs->fits, &imgtype, &status);
	return (CmpackBitpix)imgtype;
}


/* Get image data range*/
int fits_getrange(tHandle handle, double *minvalue, double *maxvalue)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int status;
	long imgtype;
	double bscale, bzero, min_val, max_val;

	status = 0;
    ffgkyj(fs->fits, "BITPIX", &imgtype, NULL, &status);
	if (status!=0)
		imgtype = 0;
    status = 0;
    ffgkyd(fs->fits, "BSCALE", &bscale, NULL, &status);
    if (status!=0)
       bscale = 1.0;
    status = 0;
    ffgkyd(fs->fits, "BZERO", &bzero, NULL, &status);
    if (status!=0)
       bzero = 0.0;

	switch (imgtype)
	{
	case BYTE_IMG:   
		min_val = 0.0;
		max_val = UCHAR_MAX;
		break;
	case SHORT_IMG:
		min_val = SHRT_MIN;
		max_val = SHRT_MAX;
		break;
	case LONG_IMG:
		min_val = LONG_MIN;
		max_val = LONG_MAX;
		break;
	case FLOAT_IMG:
		min_val = FLT_MIN;
		max_val = FLT_MAX;
		break;
	case DOUBLE_IMG:
		min_val = DBL_MIN;
		max_val = DBL_MAX;
		break;
	default:
		min_val = 0.0;
		max_val = 0.0;
	}
	if (bscale >= 0.0) {
		if (minvalue)
			*minvalue = bzero + bscale * min_val;
		if (maxvalue)
			*maxvalue = bzero + bscale * max_val;
	} else {
		if (maxvalue)
			*maxvalue = bzero + bscale * min_val;
		if (minvalue)
			*minvalue = bzero + bscale * max_val;
	}
	return 0;
}


/* Get format identifier */
char *fits_getmagic(tHandle handle)
{
	return cmpack_strdup("FITS (Flexible Image Transport System)");
}

/* Get optical filter name */
char *fits_getfilter(tHandle handle)
{
	int status;
	char value[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits, "FILTER", value, NULL, &status);
	if (status==0)
		return cmpack_strdup(value);
	status = 0;
	ffgkys(fs->fits, "FILTERS", value, NULL, &status);
	if (status==0) 
		return cmpack_strdup(value);

	return NULL;
}

/* Get object's designation */
char *fits_getobject(tHandle handle)
{
	int status;
	char object[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits, "OBJECT", object, NULL, &status);
	if (status==0)
		return cmpack_strdup(object);

	return NULL;
}

/* Reads right ascension */
char *fits_getobjra(tHandle handle)
{
	int status;
	double val;
	char rastr[FLEN_VALUE], commstr[FLEN_COMMENT], *endptr;
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits,"RA", rastr, commstr, &status);
    if (status!=0) { 
	    status = 0;
	    ffgkys(fs->fits, "OBJCTRA", rastr, commstr, &status);
    }
	if (status==0 && strlen(rastr)>0) {
		/* Convert from degrees to hours */
		if (!strchr(rastr, ':') && !strchr(rastr, ' ') && strchr(rastr, '.')) {
			val = strtod(rastr, &endptr);
			if (endptr!=rastr && val>=0.0 && val<=360.0 && strstr(commstr, "[deg]")) {
				val /= 15.0;
				sprintf(rastr, "%.8f", val);
			}
		}
		return cmpack_strdup(rastr);
	}
	return NULL;
}

/* Reads declination */
char *fits_getobjdec(tHandle handle)
{
	int status;
	char decstr[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits,"DEC", decstr, NULL, &status);
    if (status!=0) { 
	    status = 0;
	    ffgkys(fs->fits, "OBJCTDEC", decstr, NULL, &status);
    }
	if (status==0 && strlen(decstr)>0)
		return cmpack_strdup(decstr);
	return NULL;
}

/* Get observer's name */
char *fits_getobserver(tHandle handle)
{
	int status;
	char value[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits, "OBSERVER", value, NULL, &status);
	if (status==0 && strlen(value)>0)
		return cmpack_strdup(value);
	return NULL;
}

/* Get observer's name */
char *fits_getobservatory(tHandle handle)
{
	int status;
	char value[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits, "OBSERVAT", value, NULL, &status);
	if (status==0 && strlen(value)>0)
		return cmpack_strdup(value);
	return NULL;
}

/* Reads longitude */
char *fits_getobslon(tHandle handle)
{
	int status;
	char lonstr[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits,"LONGITUD", lonstr, NULL, &status);
    if (status!=0) { 
	    status = 0;
	    ffgkys(fs->fits, "GEOLON", lonstr, NULL, &status);
    }
	if (status==0 && strlen(lonstr)>0)
		return cmpack_strdup(lonstr);
	return NULL;
}

/* Reads latitude */
char *fits_getobslat(tHandle handle)
{
	int status;
	char latstr[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkys(fs->fits, "LATITUDE", latstr, NULL, &status);
    if (status!=0) { 
	    status = 0;
	    ffgkys(fs->fits, "GEOLAT", latstr, NULL, &status);
    }
	if (status==0 && strlen(latstr)>0)
		return cmpack_strdup(latstr);
	return NULL;
}

/* Reads floating point value from header */
int fits_gkyd(tHandle handle, const char *key, double *val)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int status = 0;
	
	ffgkyd(fs->fits, (char*) key, val, NULL, &status);
	if (status==0)
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get date and time of observation */
int fits_getdatetime(tHandle handle, CmpackDateTime *dt)
{
	double jd, exp;
	int end, center;
	CmpackDateTime tmp;
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	if (dt)
		memset(dt, 0, sizeof(CmpackDateTime));

	memset(&tmp, 0, sizeof(CmpackDateTime));
	if (!getdatetime(fs->fits, &tmp, &end, &center))
		return CMPACK_ERR_KEY_NOT_FOUND;

	if (end) {
		/* Tmp contains end of exposure */
		jd = cmpack_encodejd(&tmp);
		if (jd>0 && getexptime(fs->fits, &exp)) {
			if (dt)
				cmpack_decodejd(jd - exp/86400.0, dt);
			return 0;
		}
	} else if (center) {
		/* Tmp contain center of exposure */
		jd = cmpack_encodejd(&tmp);
		if (jd>0 && getexptime(fs->fits, &exp)) {
			if (dt)
				cmpack_decodejd(jd - 0.5*exp/86400.0, dt);
			return 0;
		}
	} else {
		/* Tmp contain start of exposure */
		if (dt)
			*dt = tmp;
		return 0;
	}

	return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get exposure duration */
int fits_getexptime(tHandle handle, double *val)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	if (getexptime(fs->fits, val))
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get CCD temperature */
int fits_getccdtemp(tHandle handle, double *temp)
{
	int status;
	double value;
	char buf[FLEN_VALUE];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkyd(fs->fits, "TEMPERAT", &value, NULL, &status);
	if (status==0) {
		if (temp)
			*temp = value;
		return 0;
	}
	status = 0;
	ffgkyd(fs->fits, "CCD-TEMP", &value, NULL, &status);
	if (status==0) {
		if (temp)
			*temp = value;
		return 0;
	}
	status = 0;
	ffgkys(fs->fits, "TEMP", buf, NULL, &status);
	if (status==0) { 
		if (temp)
			*temp = atof(buf); 
		return 0; 
	}
	return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get number of frames averaged or combined */
int fits_getframes(tHandle handle, int *frames)
{
	int status; 
	long val;
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	status = 0;
	ffgkyj(fs->fits, "FR_AVG", &val, NULL, &status);
	if (status==0) {
		if (frames)
			*frames = val;
		return 0;
	}
	return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Random access reading of (key, value, comment) triples from hedader */
int fits_gkyn(tHandle handle, int index, char **key, char **val, char **com)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	int status = 0;
	char k[FLEN_KEYWORD], v[FLEN_VALUE], c[FLEN_COMMENT];

	ffgkyn(fs->fits, index, k, v, c, &status);
	if (status==0) {
		if (key) *key = cmpack_strdup(k);
		if (val) *val = cmpack_strdup(v);
		if (com) *com = cmpack_strdup(c);
		return 0;
	} else {
		if (key) *key = NULL;
		if (val) *val = NULL;
		if (com) *com = NULL;
		return CMPACK_ERR_KEY_NOT_FOUND;
	}
}

/* Read image data */
int fits_getimage(tHandle handle, CmpackBitpix bitpix, int flipx, int flipy, void *buf)
{
	int status = 0, anyf;
	long naxes[2];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffgisz(fs->fits, 2, naxes, &status);
	switch (bitpix)
	{
	case CMPACK_BITPIX_SSHORT:
		ffg2di(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &status);
		return (status==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_USHORT:
		ffg2dui(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &status);
		return (status==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_SLONG:
		ffg2dj(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &status);
		return (status==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_ULONG:
		ffg2duj(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &status);
		return (status==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_FLOAT:
		ffg2de(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &status);
		return (status==0 ? 0 : CMPACK_ERR_READ_ERROR);
	case CMPACK_BITPIX_DOUBLE:
		ffg2dd(fs->fits, 1, 0, naxes[0], naxes[0], naxes[1], buf, &anyf, &status);
		return (status==0 ? 0 : CMPACK_ERR_READ_ERROR);
	default:
		return CMPACK_ERR_READ_ERROR;
	}
}

/* Put image */
int fits_putimage(tHandle handle, CmpackBitpix bitpix, const void *buf)
{
	int status = 0;
	long naxes[2];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffgisz(fs->fits, 2, naxes, &status);
	switch (bitpix)
	{
	case CMPACK_BITPIX_SSHORT:
		ffp2di(fs->fits, 1, naxes[0], naxes[0], naxes[1], (short*)buf, &status);
		return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
	case CMPACK_BITPIX_USHORT:
		ffp2dui(fs->fits, 1, naxes[0], naxes[0], naxes[1], (unsigned short*)buf, &status);
		return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
	case CMPACK_BITPIX_SLONG:
		ffp2dj(fs->fits, 1, naxes[0], naxes[0], naxes[1], (long*)buf, &status);
		return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
	case CMPACK_BITPIX_ULONG:
		ffp2duj(fs->fits, 1, naxes[0], naxes[0], naxes[1], (unsigned long*)buf, &status);
		return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
	case CMPACK_BITPIX_FLOAT:
		ffp2de(fs->fits, 1, naxes[0], naxes[0], naxes[1], (float*)buf, &status);
		return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
	case CMPACK_BITPIX_DOUBLE:
		ffp2dd(fs->fits, 1, naxes[0], naxes[0], naxes[1], (double*)buf, &status);
		return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
	default:
		return CMPACK_ERR_WRITE_ERROR;
	}
}

/* Add string to history */
int fits_puthistory(tHandle handle, const char *text)
{
	int status = 0;

	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	ffphis(fs->fits, text, &status);
	return (status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
}

/* Add string to history */
void fits_pkys(tHandle handle, const char *keyname, const char *value, const char *comment)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	ffukys(fs->fits, (char*)keyname, (char*)value, (char*)comment, &fs->status);
}
void fits_pkyi(tHandle handle, const char *keyname, int value, const char *comment)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	ffukyj(fs->fits, (char*)keyname, value, (char*)comment, &fs->status);
}
void fits_pkyl(tHandle handle, const char *keyname, int value, const char *comment)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	ffukyl(fs->fits, (char*)keyname, value, (char*)comment, &fs->status);
}
void fits_pkyd(tHandle handle, const char *keyname, double value, int prec, const char *comment)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	ffukyd(fs->fits, (char*)keyname, value, prec, (char*)comment, &fs->status);
}

/* Copy frame header, skip FITS specific keywords */
int fits_prepare(tHandle handle, int width, int height, CmpackBitpix bitpix)
{
	long naxes[2];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	naxes[0] = width;
	naxes[1] = height;
	ffphps(fs->fits, bitpix, 2, naxes, &fs->status);
	return (fs->status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
}

/* Copy image */
static int fits_fits_ui(fitsfile *handle, fitsfile *fd, int nx, int ny, int bitpix, 
						int flipx, int flipy, const CmpackBorder *border, int *sstat, int *dstat)
{
	int	anyf, j, i, len, y;
	unsigned short *image, *temp;

	/* Read image data */
	image = (unsigned short*) cmpack_calloc(nx*ny, sizeof(unsigned short));
	ffg2dui(handle,1,0,nx,nx,ny,image,&anyf,sstat);
	if (*sstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Image transformations */
	if (flipx || flipy) {
		temp = (unsigned short*) cmpack_calloc(nx*ny, sizeof(unsigned short));
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				temp[j*nx+i] = image[j*nx+i];
			}
		}
		/* Swap images */
		cmpack_free(image);
		image = temp;
	}

	/* Clear border */
	if (border) {
		if (border->left>0) {
			len = (border->left < nx ? border->left : nx);
			for (y=0; y<ny; y++)
				memset(image+(y*nx), 0, len*sizeof(unsigned short));
		}
		if (border->right) {
			len = (border->right < nx ? border->right: nx);
			for (y=0; y<ny; y++)
				memset(image+((y+1)*nx-len), 0, len*sizeof(unsigned short));
		}
		if (border->top>0) {
			len = (border->top < ny ? border->top : ny);
			for (y=0; y<len; y++)
				memset(image+(y*nx), 0, nx*sizeof(unsigned short));
		}
		if (border->bottom>0) {
			len = (border->bottom < ny ? border->bottom : ny);
			for (y=0; y<len; y++)
				memset(image+((ny-y-1)*nx), 0, nx*sizeof(unsigned short));
		}
	}

	/* Write image data */
	ffp2dui(fd,1,nx,nx,ny,image,dstat);
	if (*dstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

static int fits_fits_i(fitsfile *handle, fitsfile *fd, int nx, int ny, int bitpix, 
					   int flipx, int flipy, const CmpackBorder *border, int *sstat, int *dstat)
{
	int	anyf, j, i, len, y;
	short *image, *temp;

	/* Read image data */
	image = (short*) cmpack_calloc(nx*ny, sizeof(short));
	ffg2di(handle,1,0,nx,nx,ny,image,&anyf,sstat);
	if (*sstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Image transformations */
	if (flipx || flipy) {
		temp = (short*) cmpack_calloc(nx*ny, sizeof(short));
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				temp[j*nx+i] = image[j*nx+i];
			}
		}
		/* Swap images */
		cmpack_free(image);
		image = temp;
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(short));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(short));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(short));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(short));
	}

	/* Write image data */
	ffp2di(fd,1,nx,nx,ny,image,dstat);
	if (*dstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

/* Copy image */
static int fits_fits_uj(fitsfile *handle, fitsfile *fd, int nx, int ny, int bitpix, 
						int flipx, int flipy, const CmpackBorder *border, int *sstat, int *dstat)
{
	int	anyf, j, i, len, y;
	unsigned long *image, *temp;

	/* Read image data */
	image = (unsigned long*) cmpack_calloc(nx*ny, sizeof(unsigned long));
	ffg2duj(handle,1,0,nx,nx,ny,image,&anyf,sstat);
	if (*sstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Image transformations */
	if (flipx || flipy) {
		temp = (unsigned long*) cmpack_calloc(nx*ny, sizeof(unsigned long));
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				temp[j*nx+i] = image[j*nx+i];
			}
		}
		/* Swap images */
		cmpack_free(image);
		image = temp;
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(unsigned long));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(unsigned long));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(unsigned long));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(unsigned long));
	}

	/* Write image data */
	ffp2duj(fd,1,nx,nx,ny,image,dstat);
	if (*dstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

static int fits_fits_j(fitsfile *handle, fitsfile *fd, int nx, int ny, int bitpix, 
					   int flipx, int flipy, const CmpackBorder *border, int *sstat, int *dstat)
{
	int	anyf, j, i, len, y;
	long *image, *temp;

	/* Read image data */
	image = (long*) cmpack_calloc(nx*ny, sizeof(long));
	ffg2dj(handle,1,0,nx,nx,ny,image,&anyf,sstat);
	if (*sstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Image transformations */
	if (flipx || flipy) {
		temp = (long*) cmpack_calloc(nx*ny, sizeof(long));
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				temp[j*nx+i] = image[j*nx+i];
			}
		}
		/* Swap images */
		cmpack_free(image);
		image = temp;
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(long));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(long));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(long));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(long));
	}

	/* Write image data */
	ffp2dj(fd,1,nx,nx,ny,image,dstat);
	if (*dstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

static int fits_fits_e(fitsfile *handle, fitsfile *fd, int nx, int ny, int bitpix,
					   int flipx, int flipy, const CmpackBorder *border, int *sstat, int *dstat)
{
	int	anyf, j, i, len, y;
	float *image, *temp;

	/* Read image data */
	image = (float*) cmpack_calloc(nx*ny, sizeof(float));
	ffg2de(handle,1,0,nx,nx,ny,image,&anyf,sstat);
	if (*sstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Image transformations */
	if (flipx || flipy) {
		temp = (float*) cmpack_calloc(nx*ny, sizeof(float));
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				temp[j*nx+i] = image[j*nx+i];
			}
		}
		/* Swap images */
		cmpack_free(image);
		image = temp;
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(float));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(float));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(float));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(float));
	}

	/* Write image data */
	ffp2de(fd,1,nx,nx,ny,image,dstat);
	if (*dstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

static int fits_fits_d(fitsfile *handle, fitsfile *fd, int nx, int ny, int bitpix, 
					   int flipx, int flipy, const CmpackBorder *border, int *sstat, int *dstat)
{
	int	anyf, j, i, len, y;
	double *image, *temp;

	/* Read image data */
	image = (double*) cmpack_calloc(nx*ny, sizeof(double));
	ffg2dd(handle,1,0,nx,nx,ny,image,&anyf,sstat);
	if (*sstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Image transformations */
	if (flipx || flipy) {
		temp = (double*) cmpack_calloc(nx*ny, sizeof(double));
		for (j=0;j<ny;j++) {
			for (i=0;i<nx;i++) {
				if (flipx) 
					i = nx-i-1;
				if (flipy) 
					j = ny-j-1;
				temp[j*nx+i] = image[j*nx+i];
			}
		}
		/* Swap images */
		cmpack_free(image);
		image = temp;
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(double));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(double));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(double));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(double));
	}

	/* Write image data */
	ffp2dd(fd,1,nx,nx,ny,image,dstat);
	if (*dstat!=0) {
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

/* Convert SBIG file to FITS file */
int fits_copy(tHandle handle, fitsfile *fits, CmpackBitpix bitpix, int flipx,
			  int flipy, const CmpackBorder *border, CmpackConsole *con)
{
	int nx, ny, res, dstat=0;
	char msg[MAXLINE];
	long naxes[2];
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	if (fs->fits==NULL) 
		return CMPACK_ERR_CLOSED_FILE;

	/* Image size */
	memset(naxes, 0, 2*sizeof(long));
	ffgisz(fs->fits,2,naxes,&fs->status);
	if (fs->status!=0) {
		if (con) {
			ffgerr(fs->status, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_READ_ERROR;
	}
	nx = naxes[0];
	ny = naxes[1];

	/* Output data format */
	if (bitpix==CMPACK_BITPIX_AUTO) 
		ffgiet(fs->fits, &bitpix, &fs->status);

	/* Set output data format */
	naxes[0] = nx; 
	naxes[1] = ny;
	ffphps(fits,bitpix,2,naxes,&dstat);
	if (dstat!=0) {
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_WRITE_ERROR;
	}

	/* Copy file header */
	res = copyheader(fs->fits, fits);
	if (res!=0) 
		return CMPACK_ERR_WRITE_ERROR;
	setorigin(fits);
	setpcdate(fits);

	/* Copy image data */
	switch (bitpix)
	{
	case CMPACK_BITPIX_SCHAR:  
	case CMPACK_BITPIX_SSHORT:
		res = fits_fits_i(fs->fits, fits, nx, ny, bitpix, flipx, flipy,
			border, &fs->status, &dstat);
		break;
	case CMPACK_BITPIX_UCHAR:  
	case CMPACK_BITPIX_USHORT:
		res = fits_fits_ui(fs->fits, fits, nx, ny, bitpix, flipx, flipy,
			border, &fs->status, &dstat);
		break;
	case CMPACK_BITPIX_SLONG:
		res = fits_fits_j(fs->fits, fits, nx, ny, bitpix, flipx, flipy,
			border, &fs->status, &dstat);
		break;
	case CMPACK_BITPIX_ULONG:
		res = fits_fits_uj(fs->fits, fits, nx, ny, bitpix, flipx, flipy,
			border, &fs->status, &dstat);
		break;
	case CMPACK_BITPIX_FLOAT:
		res = fits_fits_e(fs->fits, fits, nx, ny, bitpix, flipx, flipy,
			border, &fs->status, &dstat);
		break;
	default:
		res = fits_fits_d(fs->fits, fits, nx, ny, bitpix, flipx, flipy,
			border, &fs->status, &dstat);
		break;
	}
	if (res==CMPACK_ERR_READ_ERROR) {
		if (con) {
			ffgerr(fs->status, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_READ_ERROR;
	}
	if (res==CMPACK_ERR_WRITE_ERROR) {
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_WRITE_ERROR; 
	}
	return 0;
}

void fits_setexptime(tHandle handle, double exptime)
{
	int stat = 0;
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffukyd(fs->fits, "EXPTIME", exptime, 3, "EXPOSURE DURATION IN SECONDS", &stat);
}

/* Updates date and time in FITS file */
int fits_setdatetime(tHandle handle, const CmpackDateTime *dt)
{
	char val[FLEN_VALUE];
  	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	sprintf(val, "%04d-%02d-%02d", dt->date.year, dt->date.month, dt->date.day);
	ffukys(fs->fits,"DATE-OBS", val, "UT DATE OF START", &fs->status);

	sprintf(val, "%02d:%02d:%02d.%03d", dt->time.hour, dt->time.minute, dt->time.second, dt->time.milisecond);
	ffukys(fs->fits,"TIME-OBS", val, "UT TIME OF START", &fs->status);
	return (fs->status==0 ? 0 : CMPACK_ERR_WRITE_ERROR);
}

/* Write CCD temperature */
void fits_setccdtemp(tHandle handle, double ccdtemp)
{
	int stat = 0;
  	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffukyg(fs->fits, "TEMPERAT", ccdtemp, 2, "AVERAGE CCD TEMPERATURE", &stat);
}

/* Write filter name */
void fits_setfilter(tHandle handle, const char *filter)
{
	int status = 0;
  	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffukys(fs->fits, "FILTER", (char*)filter, "OPTICAL FILTER NAME", &status);
}

/* Write object's name */
void fits_setobject(tHandle handle, const char *object)
{
	int status = 0;
  	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffukys(fs->fits, "OBJECT", (char*)object, "", &status);
}

/* Set number of frames averaged */
void fits_setframes(tHandle handle, int nframes)
{
	int status = 0;
  	CmpackFitsFile *fs = (CmpackFitsFile*)handle;

	ffukyj(fs->fits, "FR_AVG", nframes, "NUMBER OF FRAMES AVERAGED", &status);
}

/* Write package name and version */
void fits_setorigin(tHandle handle)
{
  	setorigin(((CmpackFitsFile*)handle)->fits);
}

/* Write date of creation */
void fits_setpcdate(tHandle handle)
{
	setpcdate(((CmpackFitsFile*)handle)->fits);
}

int fits_copyheader(tHandle handle, fitsfile *out)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	return copyheader(fs->fits, out);
}

int fits_clearheader(tHandle handle)
{
	CmpackFitsFile *fs = (CmpackFitsFile*)handle;
	return clearheader(fs->fits);
}
