/**************************************************************

match.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/
#define _NO_OLDNAMES

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_catfile.h"
#include "cmpack_match.h"
#include "cmpack_phtfile.h"
#include "matfun.h"
#include "matread.h"
#include "matwrite.h"
#include "matsolve.h"
#include "matsimple.h"
#include "matstack.h"
#include "match.h"

/**********************   LOCAL FUNCTIONS   ******************************/

/* Clean context */
static void match_clean(void *ctx)
{
	CmpackMatch *lc = (CmpackMatch*)ctx;

	/* Free allocated memory */
	cmpack_free(lc->x1);
	cmpack_free(lc->y1);
	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/*********************   PUBLIC FUNCTIONS   *******************************/

/* Context initialization */
CmpackMatch *cmpack_match_init(void)
{
	CmpackMatch *lc = gc_create(sizeof(CmpackMatch));
	lc->gc.clean  = &match_clean;
	lc->maxstar   = 10;
	lc->nstar     = 5;
	lc->clip      = 2.5;
	lc->spfields  = 0;
	lc->maxoffset = 5.0;
	return lc;
}


/* Attach console */
void cmpack_match_set_console(CmpackMatch *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set max. number of stars used for matching */
void cmpack_match_set_maxstars(CmpackMatch *ctx, int nstars)
{
	ctx->maxstar = nstars;
}

/* Get max. number of stars used for matching */
int cmpack_match_get_maxstars(CmpackMatch *ctx)
{
	return ctx->maxstar;
}

/* Set number of vertices of polygons */
void cmpack_match_set_vertices(CmpackMatch *ctx, int vertices)
{
	ctx->nstar = vertices;
}

/* Get number of vertices of polygons */
int cmpack_match_get_vertices(CmpackMatch *ctx)
{
	return ctx->nstar;
}

/* Set clipping threshold */
void cmpack_match_set_threshold(CmpackMatch *ctx, double threshold)
{
	ctx->clip = threshold;
}

/* Get clipping threshold */
double cmpack_match_get_threshold(CmpackMatch *ctx)
{
	return ctx->clip;
}

/* Set matching method */
void cmpack_match_set_method(CmpackMatch *ctx, int method)
{
	ctx->spfields = method;
}

/* Get matching method */
int cmpack_match_get_method(CmpackMatch *ctx)
{
	return ctx->spfields;
}

/* Set Max. offset for 'sparse fields' method */
void cmpack_match_set_maxoffset(CmpackMatch *ctx, double maxoffset)
{
	ctx->maxoffset = maxoffset;
}

/* Get Max. offset for 'sparse fields' method */
double cmpack_match_get_maxoffset(CmpackMatch *ctx)
{
	return ctx->maxoffset;
}

/* Read reference file */
int cmpack_match_readref_pht(CmpackMatch *lc, CmpackPhtFile *pht)
{
	char msg[MAXLINE];

	/* Check source file */
	if (!pht) {
		printout(lc->con, 0, "Invalid reference file context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Read reference stars */
	ReadRef(lc, pht);  
	sprintf(msg,"No. of stars   : %d",lc->c1);
	printout(lc->con, 1, msg);

  	/* Check reference file */
	if (lc->spfields==0) {
		if (lc->c1<lc->nstar) {
	        printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	} else {
		if (lc->c1<1) {
			printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	}

  	/* Normal return */
  	return 0;
}

/* Read reference file */
int cmpack_match_readref_cat(CmpackMatch *lc, CmpackCatFile *cat)
{
	char msg[MAXLINE];

	/* Check source file */
	if (!cat) {
		printout(lc->con, 0, "Invalid reference file context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Read reference stars */
	ReadCat(lc, cat);
	sprintf(msg,"No. of stars   : %d",lc->c1);
	printout(lc->con, 1, msg);

  	/* Check reference file */
	if (lc->spfields==0) {
		if (lc->c1<lc->nstar) {
	        printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	} else {
		if (lc->c1<1) {
			printout(lc->con, 0,"Too few stars in reference file!");
			return CMPACK_ERR_FEW_POINTS_REF;
		}
	}

  	/* Normal return */
  	return 0;
}

static int match(CmpackMatch *lc, CmpackPhtFile *infile, CmpackPhtFile *outfile, int *mstars)
{
	char   msg[MAXLINE];
	int    res, max2;
	CmpackMatchFrame frame;

	if (mstars)
		*mstars = 0;

	memset(&frame, 0, sizeof(CmpackMatchFrame));
	/* Make copy of reference frame */
	frame.c1 = lc->c1;
	frame.x1 = (double*) cmpack_malloc(frame.c1*sizeof(double));
	memcpy(frame.x1, lc->x1, frame.c1*sizeof(double));
	frame.y1 = (double*) cmpack_malloc(frame.c1*sizeof(double));
	memcpy(frame.y1, lc->y1, frame.c1*sizeof(double));
	/* Alloc buffers */
	frame.idr1 = (int*) cmpack_malloc(lc->maxstar*sizeof(int));
	frame.id1 = (int*) cmpack_malloc(lc->maxstar*sizeof(int));
	frame.r1 = (double*) cmpack_malloc(lc->maxstar*sizeof(double));
	frame.idr2 = (int*) cmpack_malloc(lc->maxstar*sizeof(int));
	frame.id2 = (int*) cmpack_malloc(lc->maxstar*sizeof(int));
	frame.r2 = (double*) cmpack_malloc(lc->maxstar*sizeof(double));
	max2 = (lc->nstar*(lc->nstar-1)*(lc->nstar-2))/3+1;
	frame.dev = (double*)cmpack_malloc(max2*sizeof(double));
	frame.k2 = (double*)cmpack_malloc(max2*sizeof(double));

	StInit(&frame.stack);

	/* Check parameters */
	if (lc->spfields<0 || lc->spfields>2) 
		lc->spfields = 0;
	
	if (lc->nstar<=2) {
		printout(lc->con, 0,"Number of identification stars muse be greater than 2");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->nstar>=20) {
		printout(lc->con, 0,"Number of identification stars muse be less than 20");
		return CMPACK_ERR_INVALID_PAR;
	}	  
	if (lc->maxstar<lc->nstar) {
		printout(lc->con, 0,"Number of stars used muse be greater or equal to number of identification stars");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->maxstar>=1000) {
		printout(lc->con, 0,"Number of stars used for matching muse be less than 1000");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->clip<=0) {
		printout(lc->con, 0,"Clipping factor must be greater than zero");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->maxoffset<=0) {
		printout(lc->con, 0,"Max. position offset muse be greater than zero");
		return CMPACK_ERR_INVALID_PAR;
	}
        
	/* Read input file */
    ReadSrc(&frame, infile);
	sprintf(msg,"Number of stars: %d", frame.c2);
	printout(lc->con, 1, msg);

    /* Find coincidences */
	switch (lc->spfields) 
	{
	default:
		/* Always use standard method */
		if (frame.c1>=lc->nstar && frame.c2>=lc->nstar) {
			res = Solve(lc, &frame);
		} else {
			printout(lc->con, 0,"Too few stars in source file!");
			res = CMPACK_ERR_FEW_POINTS_SRC;
		}
	    break;
	case 1:
		/* Use standard method if we have enough stars */
		if (frame.c1>=lc->nstar && frame.c2>=lc->nstar) {
		    res = Solve(lc, &frame);
		} else
		if (frame.c1>=1 && frame.c2>=1) {
		    res = Simple(lc, &frame);
		} else {
			printout(lc->con, 0,"Too few stars in source file!");
			res = CMPACK_ERR_FEW_POINTS_SRC;
		}
		break;
	case 2:
		/* Always use method for sparse fields */
		if (frame.c1>=1 && frame.c2>=1) {
			res = Simple(lc, &frame);
		} else {
			printout(lc->con, 0,"Too few stars in source file!");
			res = CMPACK_ERR_FEW_POINTS_SRC;
		}
		break;
    }

	if (frame.mstar>0) {
		sprintf(msg,"Result         : %d matched (%.0f%%)", frame.mstar, (100.0*frame.mstar)/frame.c2);
		printout(lc->con, 1,msg);
	} else {
		printout(lc->con, 1,"Result     : Coincidences not found!");
	}

    /* Write output file */
	if (res==0) {
		MatWrite(lc, &frame, outfile);
		if (mstars)
			*mstars = frame.mstar;
	}

    /* Normal return */
	StClear(&frame.stack);
	cmpack_free(frame.x1);
	cmpack_free(frame.y1);
	cmpack_free(frame.i2);
	cmpack_free(frame.x2);
	cmpack_free(frame.y2);
	cmpack_free(frame.idr1);
	cmpack_free(frame.idr2);
	cmpack_free(frame.id1);
	cmpack_free(frame.id2);
	cmpack_free(frame.r1);
	cmpack_free(frame.r2);
	cmpack_free(frame.k2);
	cmpack_free(frame.dev);
    return res;
}

int cmpack_match(CmpackMatch *lc, CmpackPhtFile *file, int *mstars)
{
	if (!file) {
		printout(lc->con, 0,"Invalid photometry file context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	return match(lc, file, file, mstars);
}

int cmpack_match_ex(CmpackMatch *lc, CmpackPhtFile *infile, 
					CmpackPhtFile *outfile, int *mstars)
{
	if (!infile) {
		printout(lc->con, 0,"Invalid source file context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (!outfile) {
		printout(lc->con, 0,"Invalid target file context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	cmpack_pht_copy(outfile, infile);
	return match(lc, infile, outfile, mstars);
}
