/**************************************************************

readall.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_phtfile.h"
#include "cmpack_allfile.h"
#include "cmpack_readall.h"

/******************* Constants and definitions **********************/

#define ALLOC_BY 64

/******************   Private data types   **************************/

/* Star data */
struct _CmpackReadAllStar
{
	int id;								/**< Star identifier (starts by 1) */
	int mask;							/**< Valid fields */
	double mag, err;					/**< Magnitude, std. error */
	struct _CmpackReadAllStar *next;	/**< Next item */
};
typedef struct _CmpackReadAllStar CmpackReadAllStar;

/* Read-all file context */
struct _CmpackReadAll
{
	CmpackContext gc;					/**< Global context */
	CmpackConsole *con;					/**< Output console */
	int aperture;						/**< Aperture index */
	int first;							/**< First frame */
	CmpackAllFile *out;					/**< Output all-file context */
};

/***********************   LOCAL FUNCTIONS   **************************/

/* Frees allocated memory */
static void readall_clean(void *ctx)
{
	CmpackReadAll *lc = (CmpackReadAll*)ctx;

	cmpack_unref(lc->con);
	gc_clean(ctx);
}
/********************   PUBLIC FUNCTIONS   ********************************/

/* Initialize the configuration context */
CmpackReadAll *cmpack_readall_init(void)
{
	CmpackReadAll *lc = gc_create(sizeof(CmpackReadAll));
	lc->gc.clean = readall_clean;
	lc->aperture = 1;
	return lc;
}

/* Attach console */
void cmpack_readall_set_console(CmpackReadAll *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set output image data format */
int cmpack_readall_set_aperture(CmpackReadAll *lc, int aperture)
{
	lc->aperture = aperture;
	return 0;
}

/* Set output image data format */
int cmpack_readall_get_aperture(CmpackReadAll *lc, int *aperture)
{
	if (aperture) *aperture = lc->aperture;
	return 0;
}

/* Open output table, start reading sequence */
int cmpack_readall_open(CmpackReadAll *lc, CmpackAllFile *out)
{	
	char msg[MAXLINE];

	if (!lc || !out)
		return CMPACK_ERR_INVALID_PAR;

	if (lc->aperture<=0) {
		printout(lc->con, 0, "Invalid aperture index.");
		return CMPACK_ERR_INVALID_PAR;
	}
	sprintf(msg, "Aperture: %d", lc->aperture);
	printout(lc->con, 1, msg);
	cmpack_all_clear(out);
	cmpack_all_set_aperture(out, lc->aperture);
	lc->out = cmpack_addref(out);
	lc->first = 1;
	return 0;
}

/* Reads magnitudes and error of the selected stars and writes them to output file */
int cmpack_readall_read(CmpackReadAll *lc, CmpackPhtFile *pht)
{
	int frame, count, i, xref;
	double jd;
	const char *filter;
	CmpackPhtAperture *aper;
	CmpackPhtStar *st;
	CmpackPhtMeasurement data;
	
	/* Prints name of input file */
	if (!pht) {
		printout(lc->con, 0, "Invalid photometry file context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Read header information */
	jd = cmpack_pht_get_jd(pht);
	if (jd<=0) {
		cmpack_unref(pht);
		return CMPACK_ERR_READ_ERROR;
	}

	if (lc->first) {
		/* Set filter */
		filter = cmpack_pht_get_filter(pht);
		if (filter)
			cmpack_all_set_filter(lc->out, filter);
		lc->first = 0;
	}

	/* Get aperture */
	aper = cmpack_pht_find_aperture(pht, lc->aperture);
	if (!aper) {
		cmpack_unref(pht);
		return CMPACK_ERR_AP_NOT_FOUND;
	}

	/* Make new frame */
	frame = cmpack_all_add_frame(lc->out, jd);
	if (frame<0) {
		cmpack_unref(aper);
		cmpack_unref(pht);
		return CMPACK_ERR_WRITE_ERROR;
	}

	/* Read star data */
	count = cmpack_pht_nstar(pht);
	for (i=0; i<count; i++) {
		st = cmpack_pht_get_star(pht, i);
		xref = cmpack_pht_star_get_refid(st);
		if (xref>=0) {
			if (cmpack_pht_star_get_measurement(st, lc->aperture, &data)==0) {
				if (data.mask & CMPACK_MASK_MAG)
					cmpack_all_set(lc->out, frame, xref, data.mask, data.magnitude, data.error);
			}
		}
		cmpack_unref(st);
	}
	cmpack_unref(aper);
	return 0;
}

/* Finish reading */
int cmpack_readall_close(CmpackReadAll *lc)
{
	cmpack_unref(lc->out);
	lc->out = NULL;
	return 0;
}
