/**************************************************************

data.c (C-Munipack project)
Table data definition
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "context.h"
#include "cmpack_common.h"
#include "cmpack_table.h"
#include "tabdata.h"

/*************************   Constants  ***********************/

#define ALLOC_BY 	64

/***********************      Datatypes      ******************/

/* Table cell data */
struct _CmpackTabCell
{
	int valid;
	union _CmpackTabCellData {
		char *sValue;
		int iValue;
		double dValue;
	} data;
};
typedef struct _CmpackTabCell CmpackTabCell;

/* Table column context */
struct _CmpackTabColumn
{
	CmpackContext gc;			/**< Global context */
	char *name;					/**< Column name */
	CmpackColumnType type;		/**< Column type */
	int prec;					/**< Number of decimal places after decimal point */
};

/* Table data */
struct _CmpackTabData
{
	int nrows, ncols;			/**< Table size */
	CmpackTabCell *data;		/**< Table data */
	int cols_size;				/**< Allocated size */
	CmpackTabColumn **list;		/**< List of columns */
};

/********************   Local functions    ***********************/

static void col_clean(void *ctx);

/* Default constructor */
static CmpackTabColumn *col_create(const char *name, CmpackColumnType type, int prec)
{
	CmpackTabColumn *col = (CmpackTabColumn*)gc_create(sizeof(CmpackTabColumn));
	col->gc.clean = col_clean;
	col->name = cmpack_strdup(name);
	col->type = type;
	col->prec = prec;
	return col;
}

/* Copy constructor */
static  CmpackTabColumn *col_copy(const CmpackTabColumn *orig)
{
	CmpackTabColumn *col = (CmpackTabColumn*)gc_create(sizeof(CmpackTabColumn));
	col->gc.clean = col_clean;
	col->name = cmpack_strdup(orig->name);
	col->type = orig->type;
	col->prec = orig->prec;
	return col;
}

/* Free allocated memory */
static void col_clear(CmpackTabColumn *col)
{
	cmpack_free(col->name);
}

/* Destructor */
static void col_clean(void *ctx)
{
	col_clear((CmpackTabColumn*)ctx);
	gc_clean(ctx);
}

/* Dump column to stdout */
static void col_dump(const CmpackTabColumn *item, CmpackConsole *con)
{
	char aux[MAXLINE];

	sprintf(aux, "\t\t%s", item->name);
	printout(con, 1, aux);
}

/* Get column name */
void tab_col_setname(CmpackTabColumn *col, const char *name)
{
	cmpack_free(col->name);
	col->name = cmpack_strdup(name);
}

/********************   Public functions    ***********************/

/* Init table data */
CmpackTabData *tab_data_create(void)
{
	CmpackTabData *data = (CmpackTabData*)cmpack_malloc(sizeof(CmpackTabData));
	memset(data, 0, sizeof(CmpackTabData));
	return data;
}

/* Free allocated memory */
void tab_data_free(CmpackTabData *data)
{
	int i, j;

	if (data) {
		/* Rows */
		if (data->data) {
			for (i=0; i<data->nrows; i++) {
				for (j=0; j<data->ncols; j++) {
					if (data->list[j]->type==CMPACK_COLUMN_STRING)
						cmpack_free(data->data[j+i*data->ncols].data.sValue);
				}
			}
			cmpack_free(data->data);
		}
		/* Columns */
		if (data->list) {
			for (j=0; j<data->ncols; j++) 
				cmpack_unref(data->list[j]);
			cmpack_free(data->list);
		}
		cmpack_free(data);
	}
}

/* Clear table data */
void tab_data_clear(CmpackTabData *data)
{
	int i, j;

	/* Rows */
	if (data->data) {
		for (i=0; i<data->nrows; i++) {
			for (j=0; j<data->ncols; j++) {
				if (data->list[j]->type==CMPACK_COLUMN_STRING)
					cmpack_free(data->data[j+i*data->ncols].data.sValue);
			}
		}
		cmpack_free(data->data);
		data->data = NULL;
	}
	/* Columns */
	if (data->list) {
		for (j=0; j<data->ncols; j++) 
			cmpack_unref(data->list[j]);
		cmpack_free(data->list);
		data->list = NULL;
	}
	data->cols_size = data->ncols = data->nrows = 0;
}

/* Copy table data */
void tab_data_copy(CmpackTabData *dst, const CmpackTabData *src)
{
	int i, j;

	/* Clear existing data */
	tab_data_clear(dst);
	/* Columns */
	if (src->ncols>0) {
		dst->cols_size = dst->ncols = src->ncols;
		dst->list = (CmpackTabColumn**)cmpack_realloc(dst->list, dst->cols_size*sizeof(CmpackTabColumn*));
		for (j=0; j<src->ncols; j++)
			dst->list[j] = col_copy(src->list[j]);
	}
	/* Rows */
	dst->nrows = src->nrows;
	if (dst->nrows>0 && dst->ncols>0) {
		dst->data = (CmpackTabCell*)cmpack_calloc(dst->nrows*dst->ncols, sizeof(CmpackTabCell));
		for (i=0; i<src->nrows; i++) {
			for (j=0; j<src->ncols; j++) {
				dst->data[j+i*dst->ncols].valid = src->data[j+i*src->ncols].valid;
				switch (src->list[j]->type)
				{
				case CMPACK_COLUMN_STRING:
					dst->data[j+i*dst->ncols].data.sValue = cmpack_strdup(src->data[j+i*src->ncols].data.sValue);
					break;
				case CMPACK_COLUMN_INTEGER:
					dst->data[j+i*dst->ncols].data.iValue = src->data[j+i*src->ncols].data.iValue;
					break;
				case CMPACK_COLUMN_DOUBLE:
					dst->data[j+i*dst->ncols].data.dValue = src->data[j+i*src->ncols].data.dValue;
					break;
				}
			}
		}
	}
}

/* Dump data */
void tab_data_dump(CmpackTabData *data, CmpackConsole *con)
{
	int i, j;
	char aux[MAXLINE], buf[MAXLINE];

	printout(con, 1, "\tColumns:");
	for (j=0; j<data->ncols; j++)
		col_dump(data->list[j], con);

	printout(con, 1, "\tData:");
	for (i=0; i<data->nrows; i++) {
		for (j=0; j<data->ncols; j++) {
			if (j==0)
				strcpy(aux, "\t\t");
			else
				strcat(aux, " ");
			switch (data->list[j]->type)
			{
			case CMPACK_COLUMN_STRING:
				if (data->data[j+i*data->ncols].valid && data->data[j+i*data->ncols].data.sValue)
					strcat(aux, data->data[j+i*data->ncols].data.sValue);
				break;
			case CMPACK_COLUMN_INTEGER:
				if (data->data[j+i*data->ncols].valid) {
					sprintf(buf, "%d", data->data[j+i*data->ncols].data.iValue);
					strcat(aux, buf);
				}
				break;
			case CMPACK_COLUMN_DOUBLE:
				if (data->data[j+i*data->ncols].valid) {
					sprintf(buf, "%.*f", data->list[j]->prec, data->data[j+i*data->ncols].data.dValue);
					strcat(aux, buf);
				}
				break;
			}
		}
		printout(con, 1, aux);
	}
}

/* Write table definition to file */
void tab_data_write_cols(CmpackTabData *data, FILE *to, const char *prefix)
{
	int	j;
	
	if (prefix)
		fprintf(to, "%s ", prefix);
	for (j=0; j<data->ncols; j++) {
		if (j>0) 
			fprintf(to, " %s", data->list[j]->name);
		else
			fprintf(to, "%s", data->list[j]->name);
	}
	fprintf(to, "\n");
}

/* Write data */
void tab_data_write_data(CmpackTabData *data, FILE *to)
{
	int i, j;

	for (i=0; i<data->nrows; i++) {
		for (j=0; j<data->ncols; j++) {
			if (j>0)
				fprintf(to, " ");
			switch (data->list[j]->type)
			{
			case CMPACK_COLUMN_STRING:
				if (data->data[j+i*data->ncols].valid && data->data[j+i*data->ncols].data.sValue)
					fprintf(to, "%s", data->data[j+i*data->ncols].data.sValue);
				break;
			case CMPACK_COLUMN_INTEGER:
				if (data->data[j+i*data->ncols].valid)
					fprintf(to, "%d", data->data[j+i*data->ncols].data.iValue);
				break;
			case CMPACK_COLUMN_DOUBLE:
				if (data->data[j+i*data->ncols].valid)
					fprintf(to, "%.*f", data->list[j]->prec, data->data[j+i*data->ncols].data.dValue);
				break;
			}
		}
		fprintf(to, "\n");
	}
}

/* Append a row */
int tab_data_add_row(CmpackTabData *data)
{
	if (data->ncols>0) {
		data->data = (CmpackTabCell*)cmpack_realloc(data->data, (data->nrows+1)*data->ncols*sizeof(CmpackTabCell));
		memset(data->data + (data->nrows*data->ncols), 0, data->ncols*sizeof(CmpackTabCell));
	}
	return data->nrows++;
}

/* Insert a row */
int tab_data_insert_row(CmpackTabData *data, int index)
{
	if (index>=0 && index<data->nrows) {
		if (data->ncols>0 && data->nrows>0) {
			CmpackTabCell *aux = (CmpackTabCell*)cmpack_malloc(data->nrows*data->ncols*sizeof(CmpackTabCell));
			memcpy(aux, data->data, data->nrows*data->ncols*sizeof(CmpackTabCell));
			data->data = (CmpackTabCell*)cmpack_realloc(data->data, (data->nrows+1)*data->ncols*sizeof(CmpackTabCell));
			if (index>0)
				memcpy(data->data, aux, index*data->ncols*sizeof(CmpackTabCell));
			memset(data->data + index*data->ncols, 0, data->ncols*sizeof(CmpackTabCell));
			memcpy(data->data + (index+1)*data->ncols, aux + index*data->ncols, 
				(data->nrows-index)*data->ncols*sizeof(CmpackTabCell));
			cmpack_free(aux);
		} else if (data->ncols>0) {
			data->data = (CmpackTabCell*)cmpack_realloc(data->data, data->ncols*sizeof(CmpackTabCell));
			memset(data->data, 0, data->ncols*sizeof(CmpackTabCell));
		}
		data->nrows++;
		return 1;
	}
	return 0;
}

/* Append a column */
int tab_data_add_col(CmpackTabData *data, const char *name, CmpackColumnType type, int prec)
{
	int j, index;
	CmpackTabCell *drow;
	const CmpackTabCell *srow;

	/* Columns */
	index = data->ncols;
	if (data->ncols>=data->cols_size) {
		data->cols_size += ALLOC_BY;
		data->list = cmpack_realloc(data->list, data->cols_size*sizeof(CmpackTabColumn));
	}
	data->list[index] = col_create(name, type, prec);
	/* Rows */
	if (data->ncols>0 && data->nrows>0) {
		CmpackTabCell *aux = (CmpackTabCell*)cmpack_malloc(data->nrows*data->ncols*sizeof(CmpackTabCell));
		memcpy(aux, data->data, data->nrows*data->ncols*sizeof(CmpackTabCell));
		data->data = (CmpackTabCell*)cmpack_realloc(data->data, data->nrows*(data->ncols+1)*sizeof(CmpackTabCell));
		for (j=0; j<data->nrows; j++) {
			drow = data->data + j*(data->ncols+1);
			srow = aux + j*(data->ncols);
			memcpy(drow, srow, data->ncols*sizeof(CmpackTabCell));
			memset(drow + data->ncols, 0, sizeof(CmpackTabCell));
		}
		cmpack_free(aux);
	} else if (data->nrows>0) {
		data->data = (CmpackTabCell*)cmpack_realloc(data->data, data->nrows*sizeof(CmpackTabCell));
		memset(data->data, 0, data->nrows*sizeof(CmpackTabCell));
	}
	return data->ncols++;
}

/* Insert a column */
int tab_data_insert_col(CmpackTabData *data, int index, const char *name, CmpackColumnType type, int prec)
{
	int j;
	CmpackTabCell *drow;
	const CmpackTabCell *srow;

	if (index>=0 && index<data->ncols) {
		/* Columns */
		if (data->ncols>=data->cols_size) {
			data->cols_size += ALLOC_BY;
			data->list = cmpack_realloc(data->list, data->cols_size*sizeof(CmpackTabColumn));
		}
		memmove(data->list+(index+1), data->list+index, (data->ncols-index)*sizeof(CmpackTabColumn));
		data->list[index] = col_create(name, type, prec);
		/* Rows */
		if (data->ncols>0 && data->nrows>0) {
			CmpackTabCell *aux = (CmpackTabCell*)cmpack_malloc(data->nrows*data->ncols*sizeof(CmpackTabCell));
			memcpy(aux, data->data, data->nrows*data->ncols*sizeof(CmpackTabCell));
			data->data = (CmpackTabCell*)cmpack_realloc(data->data, data->nrows*(data->ncols+1)*sizeof(CmpackTabCell));
			for (j=0; j<data->nrows; j++) {
				drow = data->data + j*(data->ncols+1);
				srow = aux + j*(data->ncols);
				if (index>0)
					memcpy(drow, srow, index*sizeof(CmpackTabCell));
				memset(drow + index, 0, sizeof(CmpackTabCell));
				memcpy(drow + (index+1), srow + index, (data->ncols-index)*sizeof(CmpackTabCell));
			}
			cmpack_free(aux);
		} else if (data->nrows>0) {
			data->data = (CmpackTabCell*)cmpack_realloc(data->data, data->nrows*sizeof(CmpackTabCell));
			memset(data->data, 0, data->nrows*sizeof(CmpackTabCell));
		}
		data->ncols++;
		return 1;
	}
	return 0;
}

void tab_data_set(CmpackTabData *data, int col, int row, const char *value)
{
	if (col>=0 && col<data->ncols && row>=0 && row<data->nrows) {
		CmpackTabCell *item = data->data + (col+row*data->ncols);
		switch (data->list[col]->type)
		{
		case CMPACK_COLUMN_STRING:
			if (item->data.sValue)
				cmpack_free(item->data.sValue);
			if (value) {
				item->data.sValue = cmpack_strdup(value);
				item->valid = 1;
			} else {
				item->data.sValue = NULL;
				item->valid = 0;
			}	
			break;
		case CMPACK_COLUMN_INTEGER:
			if (value) {
				item->data.iValue = atoi(value);
				item->valid = 1;
			} else {
				item->data.iValue = 0;
				item->valid = 0;
			}
			break;
		case CMPACK_COLUMN_DOUBLE:
			if (value) {
				item->data.dValue = atof(value);
				item->valid = 1;
			} else {
				item->data.dValue = 0.0;
				item->valid = 0;
			}
			break;
		}
	}
}

CmpackTabColumn *tab_data_get_col(CmpackTabData *data, int index)
{
	if (data) {
		if (index>=0 && index<data->ncols)
			return data->list[index];
	}
	return NULL;
}

int tab_data_find_col(CmpackTabData *data, const char *name)
{
	int i;

	if (data) {
		for (i=0; i<data->ncols; i++) {
			if (strcmp(data->list[i]->name, name)==0)
				return i;
		}
	}
	return -1;
}

void tab_data_seti(CmpackTabData *data, int col, int row, int value)
{
	char buf[64];

	if (col>=0 && col<data->ncols && row>=0 && row<data->nrows) {
		CmpackTabCell *item = data->data + (col+row*data->ncols);
		switch (data->list[col]->type)
		{
		case CMPACK_COLUMN_STRING:
			if (item->data.sValue)
				cmpack_free(item->data.sValue);
			sprintf(buf, "%d", value);
			item->data.sValue = cmpack_strdup(buf);
			item->valid = 1;
			break;
		case CMPACK_COLUMN_INTEGER:
			item->data.iValue = value;
			item->valid = 1;
			break;
		case CMPACK_COLUMN_DOUBLE:
			item->data.dValue = value;
			item->valid = 1;
			break;
		}
	}
}

void tab_data_setd(CmpackTabData *data, int col, int row, double value)
{
	char buf[64];

	if (col>=0 && col<data->ncols && row>=0 && row<data->nrows) {
		CmpackTabCell *item = data->data + (col+row*data->ncols);
		switch (data->list[col]->type)
		{
		case CMPACK_COLUMN_STRING:
			if (item->data.sValue)
				cmpack_free(item->data.sValue);
			sprintf(buf, "%.8f", value);
			item->data.sValue = cmpack_strdup(buf);
			item->valid = 1;
			break;
		case CMPACK_COLUMN_INTEGER:
			item->data.iValue = (int)value;
			item->valid = 1;
			break;
		case CMPACK_COLUMN_DOUBLE:
			item->data.dValue = value;
			item->valid = 1;
			break;
		}
	}
}

/* Get individual table data */
int tab_data_get(CmpackTabData *data, int col, int row, char **value)
{
	char buf[64];

	if (col>=0 && col<data->ncols && row>=0 && row<data->nrows) {
		CmpackTabCell *item = data->data + (col+row*data->ncols);
		switch (data->list[col]->type)
		{
		case CMPACK_COLUMN_STRING:
			if (item->valid && item->data.sValue) {
				if (value)
					*value = cmpack_strdup(item->data.sValue);
				return 1;
			}
			break;

		case CMPACK_COLUMN_INTEGER:
			if (item->valid) {
				if (value) {
					sprintf(buf, "%d", item->data.iValue);
					*value = cmpack_strdup(buf);
				}
				return 1;
			}
			break;

		case CMPACK_COLUMN_DOUBLE:
			if (item->valid) {
				if (value) {
					sprintf(buf, "%.8f", item->data.dValue);
					*value = cmpack_strdup(buf);
				}
				return 1;
			}
			break;
		}
	}
	return 0;
}

/* Get individual table data */
int tab_data_geti(CmpackTabData *data, int col, int row, int *value)
{
	if (col>=0 && col<data->ncols && row>=0 && row<data->nrows) {
		CmpackTabCell *item = data->data + (col+row*data->ncols);
		switch (data->list[col]->type)
		{
		case CMPACK_COLUMN_STRING:
			if (item->valid && item->data.sValue) {
				if (value)
					*value = atoi(item->data.sValue);
				return 1;
			}
			break;
			
		case CMPACK_COLUMN_INTEGER:
			if (item->valid) {
				if (value)
					*value = item->data.iValue;
				return 1;
			}
			break;
			
		case CMPACK_COLUMN_DOUBLE:
			if (item->valid) {
				if (value)
					*value = (int)item->data.dValue;
				return 1;
			}
			break;
		}
	}
	return 0;
}

/* Get individual table data */
int tab_data_getd(CmpackTabData *data, int col, int row, double *value)
{
	if (col>=0 && col<data->ncols && row>=0 && row<data->nrows) {
		CmpackTabCell *item = data->data + (col+row*data->ncols);
		switch (data->list[col]->type)
		{
		case CMPACK_COLUMN_STRING:
			if (item->valid && item->data.sValue) {
				if (value)
					*value = atof(item->data.sValue);
				return 1;
			}
			break;
			
		case CMPACK_COLUMN_INTEGER:
			if (item->valid) {
				if (value)
					*value = item->data.iValue;
				return 1;
			}
			break;
			
		case CMPACK_COLUMN_DOUBLE:
			if (item->valid) {
				if (value)
					*value = item->data.dValue;
				return 1;
			}
			break;
		}
	}
	return 0;
}

/* Get number of columns */
int tab_data_ncols(CmpackTabData *data)
{
	if (data)
		return data->ncols;
	return 0;
}

/* Get number of rows */
int tab_data_nrows(CmpackTabData *data)
{
	if (data)
		return data->nrows;
	return 0;
}

/* Get column name */
const char *tab_col_name(CmpackTabColumn *col)
{
	if (col)
		return col->name;
	return NULL;
}

/* Get column by name */
const char *tab_cols_name(CmpackTabData *data, int index)
{
	CmpackTabColumn *col = tab_data_get_col(data, index);
	if (col) 
		return col->name;
	return NULL;
}

/* Get column name */
CmpackColumnType tab_col_type(CmpackTabColumn *col)
{
	if (col)
		return col->type;
	return CMPACK_COLUMN_STRING;
}
