/**************************************************************

aperture_dlg.cpp (C-Munipack project)
Aperture selection dialog
Copyright (C) 2009 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <cmunipack.h>

#include "project.h"
#include "graph.h"
#include "aperture_dlg.h"
#include "main.h"
#include "utils.h"
#include "configuration.h"
#include "progress_dlg.h"
#include "proc_classes.h"

//-------------------------   LIST COLUMNS   --------------------------------

struct tTreeViewColumn {
	const char *caption;		// Column name
	int column;					// Model column index
	GtkTreeCellDataFunc datafn;	// Data function
	gfloat align;				// Text alignment
};

enum tApertureColumnId {
	APER_ID,					// Aperture identifier
	APER_RADIUS,				// Aperture radius
	APER_NCOLS					// Number of columns
};

enum tDataColumnId {
	DATA_COLUMN,				// Column ordinal number
	DATA_NAME,					// Column name
	DATA_NCOLS					// Number of columns
};

static void GetApertureID(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
	GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
	int id;
	double radius;
	char buf[64];

	gtk_tree_model_get(tree_model, iter, APER_ID, &id, APER_RADIUS, &radius, -1);
	g_snprintf(buf, sizeof(buf), "#%d (%.5f)", id, radius); 
	g_object_set(cell, "text", buf, NULL);
}

const static tTreeViewColumn ApertureListColumns[] = {
	{ "Aperture", 0, GetApertureID },
	{ NULL }
};

const static tTreeViewColumn DataListColumns[] = {
	{ "Data", DATA_NAME },
	{ NULL }
};

//-------------------------   POPUP MENU   ---------------------------

enum tPopupCommand
{
	CMD_SET_APERTURE
};

static const CPopupMenu::tPopupMenuItem SelectMenu[] = {
	{ CPopupMenu::MB_ITEM, CMD_SET_APERTURE, "_Select aperture" },
	{ CPopupMenu::MB_END }
};

//-------------------------   CHOOSE STARS DIALOG   --------------------------------

CApertureDlg::CApertureDlg(GtkWindow *pParent):m_LastIndex(-1), m_AperId(0), m_Row(-1),
	m_Column(0)
{
	int i;
	GtkWidget *hbox, *frame, *vbox, *scrwnd;

	// Dialog with buttons
	m_pDlg = gtk_dialog_new_with_buttons("Choose aperture", pParent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR),
		GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, NULL);

	// Dialog icon
	gchar *icon = get_icon_file("aperture");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);

	// Preview box
	hbox = gtk_hbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(m_pDlg)->vbox), hbox, TRUE, TRUE, 0);

	vbox = gtk_vbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, TRUE, 0);

	// List of apertures
	m_AperList = gtk_list_store_new(APER_NCOLS, G_TYPE_INT, G_TYPE_DOUBLE);
	m_AperView = gtk_tree_view_new_with_model(GTK_TREE_MODEL(m_AperList));
	for (i=0; ApertureListColumns[i].caption!=NULL; i++) {
		GtkTreeViewColumn *col = gtk_tree_view_column_new();
		// Set column name and alignment
		gtk_tree_view_column_set_title(col, ApertureListColumns[i].caption);
		gtk_tree_view_append_column(GTK_TREE_VIEW(m_AperView), col);
		// Add text renderer
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(col, renderer, TRUE);
		g_object_set(renderer, "xalign", ApertureListColumns[i].align, NULL);
		if (ApertureListColumns[i].datafn)
			gtk_tree_view_column_set_cell_data_func(col, renderer, ApertureListColumns[i].datafn, NULL, NULL);
		else
			gtk_tree_view_column_add_attribute(col, renderer, "text", ApertureListColumns[i].column);
	}
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(m_AperView)), GTK_SELECTION_SINGLE);
	gtk_widget_set_size_request(m_AperView, 160, 280);

	// Register callback for selection change
	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_AperView));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
	g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(selection_changed), this);

	// Make scrolled window
	scrwnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrwnd),
		GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrwnd), 
		GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(scrwnd), m_AperView);
	gtk_box_pack_start(GTK_BOX(vbox), scrwnd, TRUE, TRUE, 0);

	// List of data columns
	m_DataList = gtk_list_store_new(DATA_NCOLS, G_TYPE_INT, G_TYPE_STRING);
	m_DataView = gtk_tree_view_new_with_model(GTK_TREE_MODEL(m_DataList));
	for (i=0; DataListColumns[i].caption!=NULL; i++) {
		GtkTreeViewColumn *col = gtk_tree_view_column_new();
		// Set column name and alignment
		gtk_tree_view_column_set_title(col, DataListColumns[i].caption);
		gtk_tree_view_append_column(GTK_TREE_VIEW(m_DataView), col);
		// Add text renderer
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(col, renderer, TRUE);
		g_object_set(renderer, "xalign", DataListColumns[i].align, NULL);
		if (DataListColumns[i].datafn)
			gtk_tree_view_column_set_cell_data_func(col, renderer, DataListColumns[i].datafn, NULL, NULL);
		else
			gtk_tree_view_column_add_attribute(col, renderer, "text", DataListColumns[i].column);
	}
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(m_DataView)), GTK_SELECTION_SINGLE);
	gtk_widget_set_size_request(m_DataView, 0, 120);

	// Register callback for selection change
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_DataView));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
	g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(selection_changed), this);

	// Make scrolled window
	scrwnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrwnd),
		GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrwnd), 
		GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(scrwnd), m_DataView);
	gtk_box_pack_start(GTK_BOX(vbox), scrwnd, FALSE, TRUE, 0);

	// Graph
	frame = gtk_frame_new(NULL);
	gtk_box_pack_start(GTK_BOX(hbox), frame, true, true, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
	gtk_container_add(GTK_CONTAINER(frame), GTK_WIDGET(g_object_ref(m_Graph.Handle())));
	gtk_widget_set_size_request(m_Graph.Handle(), 600, 0);
	m_Graph.SetData(&m_Data);
	m_Graph.SetGraphPadding(2, 2, 48, 2);
	m_Graph.SetXParams("Aperture", 1, MAX_APERTURES, MAX_APERTURES, CGraph::NO_GRID, 0);
	m_Graph.SetYParams("std-dev", 0.0, 99.0, 0.1, CGraph::NO_GRID, 1);
	m_Graph.ClearColumns();
	m_Graph.AddColumn(0, -1, 0);
	m_Graph.EnableTracking(true);
	m_Graph.RegisterCallback(GraphClick, this);

	// Popup menu
	m_SelectMenu.Create(SelectMenu);
	m_SelectMenu.RegisterCallback(MenuClick, this);

	gtk_widget_show_all(hbox);
}


//
// Destructor
//
CApertureDlg::~CApertureDlg()
{
	gtk_widget_destroy(m_pDlg);
	g_object_unref(m_AperList);
	g_object_unref(m_DataList);
}


// 
// Execute the dialog
//
bool CApertureDlg::Execute()
{
	bool retval;
	int i, res;
	double radius;
	GtkWindow *pMainWnd = Global.pMainWnd->GetHandle();
	GtkTreeIter iter;
	const CSelection *sel;
	GSList *ptr;

	CProgressDlg::InitBuffer();
	m_Data.Clear();
	m_FileList = m_ColNames = NULL;

	// Load reference file
	if (!Global.pProject->GetReference(NULL, NULL)) {
		ShowError(pMainWnd, "Please, execute the matching procedure first.");
		return false;
	}

	// Check selection
	sel = Global.pProject->Selection();
	if (sel->CountStars(ST_COMPARISON)<=0 || sel->CountStars(ST_CHECK)<=0) {
		ShowError(pMainWnd, "Please, select at least one comparison star and at least one check star.");
		return false;
	}

	// Process photometry files
	gtk_tree_model_foreach(Global.pProject->FileList(), (GtkTreeModelForeachFunc)foreach_all_files, &m_FileList);
	if (!m_FileList) {
		ShowError(pMainWnd, "There are no valid frames in the list.");
		return false;
	}

	CProgressDlg pDlg(pMainWnd, "Processing files...");
	res = pDlg.Execute(ExecuteProc, this);
	if (pDlg.Cancelled()) 
		return false;
		
	if (res!=0) {
		char *msg = cmpack_formaterror(res);
		ShowError(pMainWnd, msg);
		cmpack_free(msg);
		retval = false;
	} else {
		// List of apertures
		gtk_tree_view_set_model(GTK_TREE_VIEW(m_AperView), NULL);
		gtk_list_store_clear(m_AperList);
		for (i=0; i<MAX_APERTURES; i++) {
			radius = GetApertureRadius(i+1);
			if (radius>0) {
				gtk_list_store_append(m_AperList, &iter);
				gtk_list_store_set(m_AperList, &iter, APER_ID, i+1, APER_RADIUS, radius, -1);
			}
		}
		gtk_tree_view_set_model(GTK_TREE_VIEW(m_AperView), GTK_TREE_MODEL(m_AperList));
		// List of data columns
		gtk_tree_view_set_model(GTK_TREE_VIEW(m_DataView), NULL);
		gtk_list_store_clear(m_DataList);
		i = 0;
		for (ptr=m_ColNames; ptr!=NULL; ptr=ptr->next) {
			gtk_list_store_append(m_DataList, &iter);
			gtk_list_store_set(m_DataList, &iter, DATA_COLUMN, i++, DATA_NAME, 
				(char*)ptr->data, -1);
		}
		gtk_tree_view_set_model(GTK_TREE_VIEW(m_DataView), GTK_TREE_MODEL(m_DataList));
		retval = true;
	}
	g_slist_foreach(m_FileList, (GFunc)gtk_tree_row_reference_free, NULL);
	g_slist_free(m_FileList);
	m_FileList = NULL;
	g_slist_foreach(m_ColNames, (GFunc)col_name_free, NULL);
	g_slist_free(m_ColNames);
	m_ColNames = NULL;
	if (!retval)
		return false;

	// Actual aperture
	m_LastIndex = -1;
	SetAperture(Global.pProject->Aperture());
	SetData(0);
	m_Graph.AutoZoom(true, true);

	// Set initial selection
	if (gtk_dialog_run(GTK_DIALOG(m_pDlg))!=GTK_RESPONSE_ACCEPT)
		return false;

	// Save new aperture
	Global.pProject->SetAperture(m_AperId);
	Global.pProject->Save();
	return true;
}


int CApertureDlg::ExecuteProc(CProgressDlg *sender, void *userdata)
{
	return ((CApertureDlg*)userdata)->ProcessFiles(sender);
}

int CApertureDlg::ProcessFiles(CProgressDlg *sender)
{
	int res, i, j, frameid, infiles, col_count, row_count;
	double aper, *yvals;
	char *temp, msg[128];
	CmpackTable *table = cmpack_tab_init(CMPACK_TABLE_APERTURES);
	GtkTreePath *path;
	CListProc list;

	infiles = 0;
	sender->SetMinMax(0, g_slist_length(m_FileList));
	sender->Print("------ Table of apertures ------");
	
	res = list.Open(sender, table, CMPACK_LIST_STDDEV);
	if (res==0) {
		for (GSList *node = m_FileList; node != NULL; node = node->next) {
			path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)node->data);
			frameid = Global.pProject->GetFrameID(path);
			temp = Global.pProject->GetPhotFileName(path);
			sender->SetFileName(temp);
			g_free(temp);
			sprintf(msg, "Frame #%d:", frameid);
			sender->Print(msg);
			sender->SetProgress(infiles++);
			res = list.Add(path);
			if (res!=0 || sender->Cancelled()) 
				break;
		}
		res = list.Close();
	}
	if (res==0) {
		if (sender->Cancelled())
			sprintf(msg, "Cancelled at the user's request");
		else
			sprintf(msg, "====== %d succeeded ======", infiles);
		sender->Print(msg);
	} else {
		char *aux = cmpack_formaterror(res);
		sender->Print(msg);
		cmpack_free(aux);
	}

	// Copy data from table to graph
	if (res==0 && !sender->Cancelled()) {
		m_Data.LockRedraw();
		m_Data.Clear();
		col_count = cmpack_tab_ncolumns(table)-1;
		row_count = cmpack_tab_nrows(table);
		if (col_count>0 && row_count>0) {
			m_Data.Init(col_count, row_count);
			yvals = (double*)g_malloc(col_count*sizeof(double));
			for (i=0; i<row_count; i++) {
				cmpack_tab_gtdd(table, i, 0, &aper);
				for (j=0; j<col_count; j++)
					cmpack_tab_gtdd(table, i, j+1, yvals+j);
				m_Data.Append(aper, yvals);
			}
		}
		m_Graph.ClearColumns();
		for (i=0; i<col_count; i++) {
			m_Graph.AddColumn(i, -1, 0);
			CmpackTabColumn *col = cmpack_tab_get_column(table, i+1);
			if (col) {
				m_ColNames = g_slist_append(m_ColNames, g_strdup(cmpack_tab_column_get_name(col)));
				cmpack_unref(col);
			}
		}
		m_Data.UnlockRedraw();
	}
	cmpack_unref(table);

	return res;
}

gboolean CApertureDlg::foreach_all_files(GtkTreeModel *model, GtkTreePath *path, 
										GtkTreeIter *iter, GSList **list)
{
	if (Global.pProject->GetState(iter) & CFILE_MATCHING) 
		*list = g_slist_append(*list, gtk_tree_row_reference_new(model, path));
	return FALSE;
}

void CApertureDlg::GraphClick(int message, int wparam, void *lparam, void *cb_data)
{
	CApertureDlg *pMe = (CApertureDlg*)cb_data;
		
	switch (message)
	{
	case CGraph::CB_MOUSE_MOVE:
		pMe->OnMouseMove();
		break;
	case CGraph::CB_LEFT_CLICK:
		pMe->OnLButtonClick((GdkEventButton*)lparam);
		break;
	}
}


void CApertureDlg::OnMouseMove(void)
{
	int index;
	double x, y;

	if (m_Graph.GetMousePos(&x, &y)) 
		index = m_Graph.FindByPos(x, y, m_Column);
	else
		index = -1;
	if (index>=0) {
		if (m_LastIndex!=index) {
			m_LastIndex = index;
			m_Graph.SetFocusedItem(index, m_Column);
		}
	} else {
		if (m_LastIndex>=0) {
			m_LastIndex = -1;
			m_Graph.SetFocusedItem(-1, -1);
		}
	}
}


void CApertureDlg::OnLButtonClick(const GdkEventButton *event)
{
	if (m_LastIndex>=0) {
		int aper = (int)m_Data.GetX(m_LastIndex);
		m_SelectMenu.Enable(CMD_SET_APERTURE, aper!=m_AperId);
		m_SelectMenu.Popup(event);
	}
}

void CApertureDlg::MenuClick(int message, int wparam, void *lparam, void *cb_data)
{
	CApertureDlg *pMe = (CApertureDlg*)cb_data;
		
	switch (message)
	{
	case CPopupMenu::CB_ACTIVATE:
		switch (wparam)
		{
		case CMD_SET_APERTURE:
			pMe->SetAperture((int)pMe->m_Data.GetX(pMe->m_LastIndex));
			break;
		}
	}
}

void CApertureDlg::SetAperture(int aper_id)
{
	int i;
	char buf[64];
	GtkTreeIter iter;

	for (i=0; i<m_Data.Count(); i++) {
		if ((int)m_Data.GetX(i) == aper_id) {
			m_AperId = aper_id;
			m_Row = i;
			break;
		}
	}

	m_Graph.DeleteAllTags();
	if (m_Row>=0 && m_Column>=0) {
		sprintf(buf, "Aperture #%d", m_AperId);
		m_Graph.SetTag(m_Column, m_Row, buf, RGB(255,0,0));
	}

	GtkTreeSelection *sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_AperView));
	if (gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(m_AperList), &iter, NULL, m_Row)) {
		gtk_tree_selection_select_iter(sel, &iter);
	}
}

void CApertureDlg::SetData(int column)
{
	char buf[64];

	m_Column = column;

	m_Graph.ShowColumn(-1, false);
	m_Graph.ShowColumn(m_Column, true); 

	m_Graph.DeleteAllTags();
	if (m_Row>=0 && m_Column>=0) {
		sprintf(buf, "Aperture #%d", m_AperId);
		m_Graph.SetTag(m_Column, m_Row, buf, RGB(255,0,0));
	}
}

void CApertureDlg::selection_changed(GtkTreeSelection *widget, CApertureDlg *pMe)
{
	pMe->OnSelectionChanged(widget);
}

void CApertureDlg::OnSelectionChanged(GtkTreeSelection *selection)
{
	int id;
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
		if (model == GTK_TREE_MODEL(m_AperList)) {
			gtk_tree_model_get(model, &iter, APER_ID, &id, -1);
			SetAperture(id);
		} else if (model == GTK_TREE_MODEL(m_DataList)) {
			gtk_tree_model_get(model, &iter, DATA_COLUMN, &id, -1);
			SetData(id);
		}
	}
}

void CApertureDlg::col_name_free(char *str, void *data)
{
	g_free(str);
}
