/**************************************************************

config.cpp (C-Munipack project)
Configuration file interface
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <stdio.h>

#include "configuration.h"
#include "utils.h"

// Static data
GKeyFile *CConfig::ms_KeyFile = NULL;
bool CConfig::ms_Modified = false;

void CConfig::Init(void)
{
	ms_KeyFile = g_key_file_new();
	Load();
}

void CConfig::Free(void)
{
	Flush();
	g_key_file_free(ms_KeyFile);
}

void CConfig::Load(void)
{
	char *cfg_path = g_build_filename(get_user_dir(), "config.ini", NULL);
	g_key_file_load_from_file(ms_KeyFile, cfg_path, 
		(GKeyFileFlags)(G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS), NULL);
	g_free(cfg_path);
	ms_Modified = false;
}

void CConfig::Save(void)
{
	gsize datalen;

	const char *user_dir = get_user_dir();
	force_directory(user_dir);

	char *cfg_path = g_build_filename(user_dir, "config.ini", NULL);
	char *data = g_key_file_to_data(ms_KeyFile, &datalen, NULL);
	GIOChannel *pFile = g_io_channel_new_file(cfg_path, "w+", NULL);
	if (pFile) {
		g_io_channel_write_chars(pFile, data, datalen, NULL, NULL);
		g_io_channel_close(pFile);
	}
	g_free(data);
	g_free(cfg_path);
	ms_Modified = false;
}

void CConfig::Flush(void)
{
	if (ms_Modified)
		Save();
}

void CConfig::SetStr(const char *group, const char *key, const char *val)
{
	g_key_file_set_string(ms_KeyFile, group, key, val);
	ms_Modified = true;
}

void CConfig::SetDbl(const char *group, const char *key, const double val)
{
	g_key_file_set_double(ms_KeyFile, group, key, val);
	ms_Modified = true;
}

void CConfig::SetInt(const char *group, const char *key, const int val)
{
	g_key_file_set_integer(ms_KeyFile, group, key, val);
	ms_Modified = true;
}

void CConfig::SetBool(const char *group, const char *key, bool val)
{
	g_key_file_set_boolean(ms_KeyFile, group, key, val);
	ms_Modified = true;
}


char *CConfig::GetStr(const char *group, const char *key, const char *defval)
{
	GError *error = NULL;
	char *res = g_key_file_get_string(ms_KeyFile, group, key, &error);
	if (!error)
		return res;
	else
		return g_strdup(defval);
}

bool CConfig::TryGetStr(const char *group, const char *key, char **value)
{
	GError *error = NULL;
	char *res = g_key_file_get_string(ms_KeyFile, group, key, &error);
	if (!error) {
		if (value) *value = g_strdup(res);
		return true;
	} else {
		if (value) *value = NULL;
		return false;
	}
}

double CConfig::GetDbl(const char *group, const char *key, double defval)
{
	GError *error = NULL;
	double res = g_key_file_get_double(ms_KeyFile, group, key, &error);
	if (!error)
		return res;
	else
		return defval;
}

bool CConfig::TryGetDbl(const char *group, const char *key, double *value)
{
	GError *error = NULL;
	double res = g_key_file_get_double(ms_KeyFile, group, key, &error);
	if (!error) {
		if (value) *value = res;
		return true;
	} else {
		if (value) *value = 0.0;
		return false;
	}
}

int CConfig::GetInt(const char *group, const char *key, const int defval)
{
	GError *error = NULL;
	int res = g_key_file_get_integer(ms_KeyFile, group, key, &error);
	if (!error)
		return res;
	else
		return defval;
}

bool CConfig::TryGetInt(const char *group, const char *key, int *value)
{
	GError *error = NULL;
	int res = g_key_file_get_integer(ms_KeyFile, group, key, &error);
	if (!error) {
		if (value) *value = res;
		return true;
	} else {
		if (value) *value = 0;
		return false;
	}
}

bool CConfig::GetBool(const char *group, const char *key, const bool defval)
{
	GError *error = NULL;
	bool res = g_key_file_get_boolean(ms_KeyFile, group, key, &error)!=0;
	if (!error)
		return res;
	else
		return defval;
}

bool CConfig::TryGetBool(const char *group, const char *key, bool *value)
{
	GError *error = NULL;
	bool res = g_key_file_get_boolean(ms_KeyFile, group, key, &error)!=0;
	if (!error) {
		if (value) *value = res;
		return true;
	} else {
		if (value) *value = false;
		return false;
	}
}
