/**************************************************************

preview.cpp (C-Munipack project)
Preview widget
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <glib.h>
#include <string.h>
#include <math.h>
#include <cmunipack.h>

#include "preview.h"
#include "utils.h"
#include "configuration.h"
#include "ccdfile_dlg.h"
#include "phtfile_dlg.h"

#define FONT_SIZE		10

//
// Chart data
//
struct tStarRec
{
	double		x, y;		// Position
	double		mag;		// Brightness in smallest aperture
	tStarType	type;		// Selection type
};


//
// Table data
//
struct tDataRec
{
	double		x, y;		// Position
};


//
// grayscale mapping
//
static unsigned long grayscale(double x)
{
	x = 255.0*x;

	if (x<=0.0)
		return 0x000000;
	if (x>=255.0)
		return 0xFFFFFF;

	int ci = (int)x;
	return ci | (ci<<8) | (ci<<16);
}

//
// Compare stars by magnitude
//
static int compare_stars(const void *x, const void *y)
{
	if (((tStarRec*)x)->mag < ((tStarRec*)y)->mag)
		return -1;
	else if (((tStarRec*)x)->mag > ((tStarRec*)y)->mag)
		return +1;
	else
		return 0;
}


// 
// Compute size of a paragraph
//
static void TextExtents(cairo_t *cr, const char *text, int *width, int *height)
{
	int w, h;
	char *buf, *start, *ptr;
	cairo_text_extents_t extents;

	w = h = 0;
	buf = g_strdup(text);
	start = buf;
	ptr = strchr(start, '\n');
	while (ptr) {
		*ptr = '\0';
		if (strlen(start)>0) {
			cairo_text_extents(cr, start, &extents);
			h += (int)(extents.height*1.5);
			if (w < (int)extents.width)
				w = (int)extents.width;
		} else {
			cairo_text_extents(cr, "X", &extents);
			h += (int)(extents.height*1.5);
		}
		start = ptr + 1;
		ptr = strchr(start, '\n');
	}
	if (strlen(start)>0) {
		cairo_text_extents(cr, start, &extents);
		h += (int)(extents.height*1.5);
		if (w < (int)extents.width)
			w = (int)extents.width;
	}
	g_free(buf);
	if (width)
		*width = w;
	if (height)
		*height = h;
}

// 
// Compute size of a paragraph
//
static void DrawText(cairo_t *cr, int x, int y, int width, int height, const char *text)
{
	char *buf, *start, *ptr;
	cairo_text_extents_t extents;

	buf = g_strdup(text);
	start = buf;
	ptr = strchr(start, '\n');
	while (ptr) {
		*ptr = '\0';
		if (strlen(start)>0) {
			cairo_text_extents(cr, start, &extents);
			cairo_move_to(cr, x + (width - (int)extents.width)/2, y + (int)extents.height);
			cairo_show_text(cr, start);
			y += (int)(extents.height*1.5);
		} else {
			cairo_text_extents(cr, "X", &extents);
			y += (int)(extents.height*1.5);
		}
		start = ptr + 1;
		ptr = strchr(start, '\n');
	}
	if (strlen(start)>0) {
		cairo_text_extents(cr, start, &extents);
		cairo_move_to(cr, x + (width - (int)extents.width)/2, y + (int)extents.height);
		cairo_show_text(cr, start);
	}
	g_free(buf);
}

//
// Constructor
//
CPreview::CPreview():m_Path(NULL), m_pParent(NULL), m_TypeMask(FILE_TYPE_ALL), 
	m_Pixmap(NULL), m_PixWidth(0), m_PixHeight(0), m_ApertureId(1), 
	m_Image(NULL), m_ImgWidth(0), m_ImgHeight(0), m_Stars(NULL), m_StCount(0), 
	m_PhtWidth(0), m_PhtHeight(0), m_Data(NULL), m_Text(NULL)
{
	GtkWidget *vbox;

	// Preview frame
	m_Widget = gtk_frame_new(NULL);
	vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(m_Widget), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);

	// Preview image
	m_DrawArea = gtk_drawing_area_new();
	gtk_widget_set_size_request(m_DrawArea, 160, 160);
	gtk_box_pack_start(GTK_BOX(vbox), m_DrawArea, FALSE, TRUE, 0);
	gtk_widget_set_events(m_DrawArea, GDK_EXPOSURE_MASK | 
			GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK |
			GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);

	// Signals used to handle backing pixmap 
	g_signal_connect(G_OBJECT(m_DrawArea), "expose_event", G_CALLBACK(expose_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "configure_event", G_CALLBACK(configure_event), this);
	g_signal_connect(G_OBJECT(m_DrawArea), "button_press_event", G_CALLBACK(button_press_event), this);

	// Frame info
	m_Info = gtk_text_view_new();
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(m_Info), GTK_WRAP_WORD_CHAR);
	gtk_text_view_set_editable(GTK_TEXT_VIEW(m_Info), false);
	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(m_Info), false);
	gtk_text_view_set_justification(GTK_TEXT_VIEW(m_Info), GTK_JUSTIFY_CENTER);
	gtk_box_pack_start(GTK_BOX(vbox), m_Info, TRUE, TRUE, 0);

	gtk_widget_show_all(m_Widget);
}


//
// Destructor
//
CPreview::~CPreview()
{
	g_free(m_Stars);
	g_free(m_Data);
	g_free(m_Text);
	if (m_Image)
		cairo_surface_destroy(m_Image);
	if (m_Pixmap)
		g_object_unref(m_Pixmap);
	g_free(m_Path);
	gtk_widget_destroy(m_Widget);
}


//
// Set type mask
//
void CPreview::SetTypeMask(int bitmask)
{
	m_TypeMask = bitmask;
}


//
// Clear all
//
void CPreview::Clear(void)
{
	// Clear image
	cairo_surface_destroy(m_Image);
	m_Image = NULL;
	m_ImgWidth = m_ImgHeight = 0;
	g_free(m_Stars);
	m_Stars = NULL;
	m_StCount = 0;
	g_free(m_Data);
	m_Data = NULL;
	m_DataCount = 0;
	g_free(m_Path);
	m_Path = NULL;
	g_free(m_Text);
	m_Text = NULL;
	// Clear infobox
	gtk_text_buffer_set_text(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_Info)), "", -1);
	// Invalidate paint box
	Redraw();
}


//
// Set file path
//
void CPreview::SetPath(const char *path)
{
	if (path) {
		int len = (int)strlen(path);
		m_Path = (char*)g_realloc(m_Path, (len+1)*sizeof(char));
		strcpy(m_Path, path);
	} else {
		g_free(m_Path);
		m_Path = NULL;
	}
}


//
// Set parent
//
void CPreview::SetParent(GtkWindow *parent)
{
	m_pParent = parent;
}


//
// Load image file
//
void CPreview::LoadImage(const char *filename)
{
	CmpackCcdFile *ccd_file;

	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_ccd_open(&ccd_file, f, CMPACK_OPEN_READONLY)==0) {
		UpdateImageData(ccd_file);
		UpdateImageInfo(ccd_file);
		cmpack_unref(ccd_file);
		SetPath(filename);
	}
	g_free(f);
}

// Load image file
void CPreview::UpdateImageInfo(CmpackCcdFile *file)
{
	double exptime, ccdtemp;
	CmpackDateTime dt;
	char msg[1024], *f2;
	const char *filter, *object, *observer;
	GtkTextBuffer *text = gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_Info));
	GtkTextIter ptr;

	// Get image info
	gtk_text_buffer_get_end_iter(text, &ptr);
	if (cmpack_ccd_get_datetime(file, &dt)==0) {
		sprintf(msg, "Date: %04d-%02d-%02d\nTime: %d:%02d:%02d",
			dt.date.year, dt.date.month, dt.date.day, 
			dt.time.hour, dt.time.minute, dt.time.second);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	} else {
		gtk_text_buffer_insert(text, &ptr, "Invalid date or time", -1);
	}
	filter = cmpack_ccd_get_filter(file);
	if (filter) {
		f2 = g_locale_to_utf8(filter, -1, NULL, NULL, NULL);
		sprintf(msg, "\nFilter: %s", f2);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
		g_free(f2);
	}
	if (cmpack_ccd_get_exptime(file, &exptime)==0) {
		sprintf(msg, "\nExp. duration: %.2f s", exptime);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
	if (cmpack_ccd_get_ccdtemp(file, &ccdtemp)==0) {
		sprintf(msg, "\nCCD temperature: %.1f \xC2\xB0""C", ccdtemp);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
	object = cmpack_ccd_get_object(file);
	if (object) {
		f2 = g_locale_to_utf8(object, -1, NULL, NULL, NULL);
		sprintf(msg, "\nObject: %s", f2);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
		g_free(f2);
	}
	observer = cmpack_ccd_get_observer(file);
	if (observer) {
		f2 = g_locale_to_utf8(observer, -1, NULL, NULL, NULL);
		sprintf(msg, "\nObserver: %s", f2);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
		g_free(f2);
	}
}


//
// Update image data
//
void CPreview::UpdateImageData(CmpackCcdFile *ccd_file)
{
	bool invert;
	double minval, maxval, datamin, datamax, val, a, *ccd;
	double scale, offset;
	int *hist, count, i, j, k, index, width, height, pixels;
	unsigned long *dptr;

	// Configuration parameters
	minval = CConfig::GetDbl("CCD", "MinValue", 0.0);
	maxval = CConfig::GetDbl("CCD", "MaxValue", 65535.0);
	invert = CConfig::GetBool("Environment", "ImageInvert");

	// Get image size
	width = height = 0;
	if (cmpack_ccd_get_size(ccd_file, &width, &height)==0) {
		// Resize image surface
		if (m_ImgHeight!=height || m_ImgWidth!=width) {
			if (m_Image) {
				cairo_surface_destroy(m_Image);
				m_Image = NULL;
			}
			m_ImgWidth = width;
			m_ImgHeight = height;
			if (m_ImgWidth>0 && m_ImgHeight>0)
				m_Image = cairo_image_surface_create(CAIRO_FORMAT_RGB24, m_ImgWidth, m_ImgHeight);
		}
		// Get CCD frame data
		pixels = width*height;
		ccd = (double*)g_malloc(pixels*sizeof(double));
		cmpack_ccd_get_image(ccd_file, CMPACK_BITPIX_DOUBLE, 0, 0, ccd);
		// Autocontrast
		datamax = minval;
		datamin = maxval;
		for (i=0; i<pixels; i++) {
			val = ccd[i];
			if (val>minval && val<maxval) {
				if (datamin > val)
					datamin = val;
				if (datamax < val)
					datamax = val;
			}
		}
		scale = offset = 0.0;
		if (((datamax-datamin)>65536*DBL_EPSILON)) {
			a = 65535.0/(datamax-datamin);
			hist = (int*)g_malloc0(65536*sizeof(int));
			for (i=0; i<pixels; i++) {
				val = ccd[i];
				if (val>minval && val<maxval) {
					val = a*(val-datamin);
					if (val<0)
						index = 0;
					else if (val>65535)
						index = 65535;
					else
						index = (int)(val+0.5);
					hist[index]++;
				}
			}
			for (j=count=0; j<65535 && count<0.01*pixels; j++)
				count += hist[j];
			for (k=65535, count=0; k>j && count<0.01*pixels; k--)
				count += hist[k];
			if (k-j>16) {
				scale  = -a/(k-j);
				offset = -scale*(k/a+datamin);
			} else {
				scale = 1.0/(datamax-datamin);
				offset = -scale*datamin;
			}
			g_free(hist);
		} else if (datamax>datamin) {
			scale = 1.0/(datamax-datamin);
			offset = -scale*datamin;
		}
		if (!invert) {
			scale = -scale;
			offset = 1.0-offset;
		}
		if (m_Image) {
			dptr = (unsigned long*)cairo_image_surface_get_data(m_Image);
			for (i=0; i<pixels; i++) {
				*dptr++ = grayscale(scale*ccd[i] + offset);
			}
		}
		g_free(ccd);
	}
}

// Load photometry file
void CPreview::LoadChart(const char *filename)
{
	CmpackPhtFile *pht;

	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_pht_load(&pht, f, CMPACK_LOAD_DEFAULT)==0) {
		UpdateChartData(pht);
		UpdateChartInfo(pht);
		SetPath(filename);
		cmpack_unref(pht);
	}
	g_free(f);
}

// Load image file
void CPreview::UpdateChartInfo(CmpackPhtFile *file)
{
	int stars;
	double jd, exptime;
	CmpackDateTime dt;
	char msg[1024], *f2;
	const char *filter;
	GtkTextBuffer *text = gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_Info));
	GtkTextIter ptr;

	// Get image info
	gtk_text_buffer_get_end_iter(text, &ptr);
	jd = cmpack_pht_get_jd(file);
	if (jd>0) {
		cmpack_decodejd(jd, &dt);
		sprintf(msg, "Date: %04d-%02d-%02d\nTime: %d:%02d:%02d",
			dt.date.year, dt.date.month, dt.date.day, 
			dt.time.hour, dt.time.minute, dt.time.second);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	} else {
		gtk_text_buffer_insert(text, &ptr, "Invalid date or time", -1);
	}
	filter = cmpack_pht_get_filter(file);
	if (filter && strlen(filter)>0) {
		f2 = g_locale_to_utf8(filter, -1, NULL, NULL, NULL);
		sprintf(msg, "\nFilter: %s", f2);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
		g_free(f2);
	}
	exptime = cmpack_pht_get_exptime(file);
	if (exptime>0) {
		sprintf(msg, "\nExp. duration: %.2f s", exptime);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
	stars = cmpack_pht_nstar(file);
	if (stars>0) {
		sprintf(msg, "\nNumber of stars: %d", stars);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
}

// Load photometry file
void CPreview::UpdateChartData(CmpackPhtFile *file)
{
	int i, count, width, height;
	double x, y, stddev, *maglist;
	CmpackPhtMeasurement mag;

	cmpack_pht_get_size(file, &width, &height);
	if (width>0 && height>0) {
		m_PhtWidth = width;
		m_PhtHeight = height;
		// Make list of stars
		count = cmpack_pht_nstar(file);
		m_Stars = (tStarRec*)g_malloc(count*sizeof(tStarRec));
		m_StCount = 0;
		for (i=0; i<count; i++) {
			CmpackPhtStar *st = cmpack_pht_get_star(file, i);
			if (cmpack_pht_star_get_center(st, &x, &y)) {
				cmpack_pht_star_get_measurement(st, m_ApertureId, &mag);
				if (mag.mask & CMPACK_MASK_MAG) {
					m_Stars[m_StCount].x = x;
					m_Stars[m_StCount].y = y;
					m_Stars[m_StCount].mag = mag.magnitude;
					m_Stars[m_StCount].type = ST_NOT_SELECTED;
					m_StCount++;
				}
			}
			cmpack_unref(st);
		}
		// Sort stars by magnitude
		if (m_StCount>1)
			qsort(m_Stars, m_StCount, sizeof(tStarRec), compare_stars);
		// Compute reference magnitude
		if (m_StCount>0) {
			maglist = (double*)g_malloc(m_StCount*sizeof(double));
			for (i=0; i<m_StCount; i++)
				maglist[i] = m_Stars[i].mag;
			cmpack_robustmean(m_StCount, maglist, &m_RefMag, &stddev);
			g_free(maglist);
		}
	}
}

// Load catalogue file
void CPreview::LoadCatalog(const char *filename)
{
	CmpackCatFile *cat;

	// Catalogue file
	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_cat_load(&cat, f, CMPACK_LOAD_DEFAULT)==0) {
		UpdateCatalogData(cat);
		UpdateCatalogInfo(cat);
		SetPath(filename);
		cmpack_unref(cat);
	}
	g_free(f);
}

// Load image file
void CPreview::UpdateCatalogInfo(CmpackCatFile *file)
{
	int stars;
	double jd, exptime;
	CmpackDateTime dt;
	char msg[1024], *f2;
	const char *filter;
	GtkTextBuffer *text = gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_Info));
	GtkTextIter ptr;

	// Get image info
	gtk_text_buffer_get_end_iter(text, &ptr);
	jd = cmpack_cat_get_jd(file);
	if (jd>0) {
		cmpack_decodejd(jd, &dt);
		sprintf(msg, "Date: %04d-%02d-%02d\nTime: %d:%02d:%02d",
			dt.date.year, dt.date.month, dt.date.day, 
			dt.time.hour, dt.time.minute, dt.time.second);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	} else {
		gtk_text_buffer_insert(text, &ptr, "Invalid date or time", -1);
	}
	filter = cmpack_cat_get_filter(file);
	if (filter && strlen(filter)>0) {
		f2 = g_locale_to_utf8(filter, -1, NULL, NULL, NULL);
		sprintf(msg, "\nFilter: %s", f2);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
		g_free(f2);
	}
	exptime = cmpack_cat_get_exptime(file);
	if (exptime>0) {
		sprintf(msg, "\nExp. duration: %.2f s", exptime);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
	stars = cmpack_cat_nstar(file);
	if (stars>0) {
		sprintf(msg, "\nNumber of stars: %d", stars);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
}


//
// Load catalogue file
//
void CPreview::UpdateCatalogData(CmpackCatFile *file)
{
	int i, id, count, width, height;
	double x, y, stddev, mag, *maglist;

	cmpack_cat_get_size(file, &width, &height);
	if (width>0 && height>0) {
		m_PhtWidth = width;
		m_PhtHeight = height;
		// Make list of stars
		count = cmpack_cat_nstar(file);
		m_Stars = (tStarRec*)g_malloc(count*sizeof(tStarRec));
		m_StCount = 0;
		for (i=0; i<count; i++) {
			CmpackCatStar *st = cmpack_cat_get_star(file, i);
			if (cmpack_cat_star_get_center(st, &x, &y) &&
				cmpack_cat_star_get_magnitude(st, &mag)) {
					m_Stars[m_StCount].x = x;
					m_Stars[m_StCount].y = y;
					m_Stars[m_StCount].mag = mag;
					id = cmpack_cat_star_get_id(st);
					switch (cmpack_cat_get_star_type(file, id))
					{
					case CMPACK_SELECT_VAR:
						m_Stars[m_StCount].type = ST_VARIABLE;
						break;
					case CMPACK_SELECT_COMP:
						m_Stars[m_StCount].type = ST_COMPARISON;
						break;
					case CMPACK_SELECT_CHECK:
						m_Stars[m_StCount].type = ST_CHECK;
						break;
					default:
						m_Stars[m_StCount].type = ST_NOT_SELECTED;
						break;
					}
					m_StCount++;
			}
			cmpack_unref(st);
		}
		// Sort stars by magnitude
		if (m_StCount>1)
			qsort(m_Stars, m_StCount, sizeof(tStarRec), compare_stars);
		// Compute reference magnitude
		if (m_StCount>0) {
			maglist = (double*)g_malloc(m_StCount*sizeof(double));
			for (i=0; i<m_StCount; i++)
				maglist[i] = m_Stars[i].mag;
			cmpack_robustmean(m_StCount, maglist, &m_RefMag, &stddev);
			g_free(maglist);
		}
	}
}


//
// Load table
//
void CPreview::LoadTable(const char *filename)
{
	CmpackTable *tab;

	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_tab_load(&tab, f, CMPACK_LOAD_DEFAULT)==0) {
		UpdateTableData(tab);
		UpdateTableInfo(tab);
		SetPath(filename);
		cmpack_unref(tab);
	}
	g_free(f);
}


//
// Update table info
//
void CPreview::UpdateTableInfo(CmpackTable *tab)
{
	int aperture;
	char msg[1024], *f2;
	const char *filter;
	GtkTextBuffer *text = gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_Info));
	GtkTextIter ptr;

	gtk_text_buffer_get_end_iter(text, &ptr);
	switch (cmpack_tab_get_type(tab))
	{
	case CMPACK_TABLE_LIGHTCURVE:
		gtk_text_buffer_insert(text, &ptr, "Light curve", -1);
		break;
	case CMPACK_TABLE_TRACKLIST:
		gtk_text_buffer_insert(text, &ptr, "Track list", -1);
		break;
	case CMPACK_TABLE_APERTURES:
		gtk_text_buffer_insert(text, &ptr, "Apertures", -1);
		break;
	case CMPACK_TABLE_MUNIFIND:
		gtk_text_buffer_insert(text, &ptr, "Munifind data", -1);
		break;
	case CMPACK_TABLE_AIRMASS:
		gtk_text_buffer_insert(text, &ptr, "Air-mass coefficients", -1);
		break;
	default:
		gtk_text_buffer_insert(text, &ptr, "Table", -1);
		break;
	}
	filter = cmpack_tab_gkys(tab, "Filter");
	if (filter && strlen(filter)>0) {
		f2 = g_locale_to_utf8(filter, -1, NULL, NULL, NULL);
		sprintf(msg, "\nFilter: %s", f2);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
		g_free(f2);
	}
	aperture = cmpack_tab_gkyi(tab, "Aperture", 0);
	if (aperture>0) {
		sprintf(msg, "\nAperture: %d", aperture);
		gtk_text_buffer_insert(text, &ptr, msg, -1);
	}
}


//
// Load table data
//
void CPreview::UpdateTableData(CmpackTable *tab)
{
	int i, j, nrows, ncols;
	double x, y;

	switch (cmpack_tab_get_type(tab))
	{
	case CMPACK_TABLE_LIGHTCURVE:
		nrows = cmpack_tab_nrows(tab);
		ncols = cmpack_tab_ncolumns(tab);
		if (nrows>0 && ncols>=2) {
			m_Data = (tDataRec*)g_malloc(nrows*sizeof(tDataRec));
			m_DataCount = 0;
			for (i=0; i<nrows; i++) {
				if (cmpack_tab_gtdd(tab, i, 0, &x) && cmpack_tab_gtdd(tab, i, 1, &y)) {
					if (x>=0 && y<=99) {
						m_Data[m_DataCount].x = x;
						m_Data[m_DataCount].y = -y;
						m_DataCount++;
					}
				}
			}
		}
		break;

	case CMPACK_TABLE_TRACKLIST:
	case CMPACK_TABLE_AIRMASS:
		nrows = cmpack_tab_nrows(tab);
		ncols = cmpack_tab_ncolumns(tab);
		if (nrows>0 && ncols>=2) {
			m_Data = (tDataRec*)g_malloc(nrows*sizeof(tDataRec));
			m_DataCount = 0;
			for (i=0; i<nrows; i++) {
				if (cmpack_tab_gtdd(tab, i, 0, &x) && cmpack_tab_gtdd(tab, i, 1, &y)) {
					if (x>=0) {
						m_Data[m_DataCount].x = x;
						m_Data[m_DataCount].y = y;
						m_DataCount++;
					}
				}
			}
		}
		break;

	case CMPACK_TABLE_APERTURES:
		nrows = cmpack_tab_nrows(tab);
		ncols = cmpack_tab_ncolumns(tab);
		if (nrows>0 && ncols>=2) {
			m_Data = (tDataRec*)g_malloc(nrows*sizeof(tDataRec));
			m_DataCount = 0;
			for (i=0; i<nrows; i++) {
				if (cmpack_tab_gtdi(tab, i, 0, &j) && cmpack_tab_gtdd(tab, i, 1, &y)) {
					if (j>0 && y<=99) {
						m_Data[m_DataCount].x = j;
						m_Data[m_DataCount].y = y;
						m_DataCount++;
					}
				}
			}
		}
		break;

	case CMPACK_TABLE_MUNIFIND:
		nrows = cmpack_tab_nrows(tab);
		ncols = cmpack_tab_ncolumns(tab);
		if (nrows>0 && ncols>=3) {
			m_Data = (tDataRec*)g_malloc(nrows*sizeof(tDataRec));
			m_DataCount = 0;
			for (i=0; i<nrows; i++) {
				if (cmpack_tab_gtdd(tab, i, 1, &x) && cmpack_tab_gtdd(tab, i, 2, &y)) {
					if (x<99) {
						m_Data[m_DataCount].x = x;
						m_Data[m_DataCount].y = y;
						m_DataCount++;
					}
				}
			}
		}
		break;

	default:
		break;
	}

	// Compute reference magnitude
	if (m_DataCount>0) {
		m_MinX = m_MaxX = m_Data[0].x;
		m_MinY = m_MaxY = m_Data[0].y;
		for (i=1; i<m_DataCount; i++) {
			x = m_Data[i].x;
			if (m_MinX > x)
				m_MinX = x;
			if (m_MaxX < x)
				m_MaxX = x;
			y = m_Data[i].y;
			if (m_MinY > y)
				m_MinY = y;
			if (m_MaxY < y)
				m_MaxY = y;
		}
	}
}

//
// Load table
//
void CPreview::LoadReadall(const char *filename)
{
	m_Text = g_strdup("Munifind data file\n(no preview available)");
	SetPath(filename);
}

/* Create a new backing pixmap of the appropriate size */
gint CPreview::configure_event(GtkWidget *widget, GdkEventConfigure *event, CPreview *pMe)
{
	pMe->OnConfigure(widget, event);
	return TRUE;
}

void CPreview::OnConfigure(GtkWidget *widget, GdkEventConfigure *event)
{
	int width, height;

	width = widget->allocation.width;
	height = widget->allocation.height;
	if (width!=m_PixWidth && height!=m_PixHeight) {
		if (m_Pixmap) {
			g_object_unref(m_Pixmap);
			m_Pixmap = NULL;
		}
		if (width>0 && height>0) {
			m_Pixmap = gdk_pixmap_new(widget->window, width, height, -1);
			m_PixWidth = width;
			m_PixHeight = height;
		} else {
			m_PixHeight = m_PixWidth = 0;
		}
		Redraw();
	}
}

void CPreview::Update(const char *filename)
{
	Clear();

	switch (FileType(filename) & m_TypeMask) 
	{
	case FILE_TYPE_IMAGE:
		LoadImage(filename);
		break;
	case FILE_TYPE_PHOT:
		LoadChart(filename);
		break;
	case FILE_TYPE_CAT:
		LoadCatalog(filename);
		break;
	case FILE_TYPE_TABLE:
		LoadTable(filename);
		break;
	case FILE_TYPE_READALL:
		LoadReadall(filename);
	default:
		break;
	}

	Redraw();
}

/* Redraw the screen from the backing pixmap */
gint CPreview::expose_event(GtkWidget *widget, GdkEventExpose *event, CPreview *pMe)
{
	pMe->OnExposeEvent(widget, event);
	return FALSE;
}

void CPreview::OnExposeEvent(GtkWidget *widget, GdkEventExpose *event)
{
	gdk_draw_drawable(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		m_Pixmap, event->area.x, event->area.y, event->area.x, event->area.y,
		event->area.width, event->area.height);
}

/* Draw a rectangle on the screen */
void CPreview::Redraw(void)
{
	double x, y, r, zoom_x, zoom_y, offset_x, offset_y, zoom;
	int i, dx, dy, width, height;
	
	if (m_Pixmap && m_PixWidth>0 && m_PixHeight>0) {
		cairo_t *cr = gdk_cairo_create(m_Pixmap);
		// Paint background
		cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
		cairo_paint(cr);

		// Draw image
		if (m_Image && m_ImgWidth>0 && m_ImgHeight>0) {
			zoom_x = (double)m_PixWidth/m_ImgWidth;
			zoom_y = (double)m_PixHeight/m_ImgHeight;
			zoom = (zoom_x < zoom_y ? zoom_x : zoom_y);
			width = (int)(zoom*m_ImgWidth+0.5);
			height = (int)(zoom*m_ImgHeight+0.5);
			dx = (int)((m_PixWidth - width)/2);
			dy = (int)((m_PixHeight - height)/2);
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
			cairo_identity_matrix(cr);
			cairo_pattern_t *pat = cairo_pattern_create_for_surface(m_Image);
			cairo_pattern_set_filter(pat, CAIRO_FILTER_GOOD);
			cairo_translate(cr, dx, dy);
			cairo_scale(cr, zoom, zoom);
			cairo_set_source(cr, pat);
			cairo_paint(cr);
			cairo_pattern_destroy(pat);
		}

		// Draw chart
		if (m_Stars && m_PhtWidth>0 && m_PhtHeight>0) {
			// Mapping coefficients
			zoom_x = (double)m_PixWidth/m_PhtWidth;
			zoom_y = (double)m_PixHeight/m_PhtHeight;
			zoom = (zoom_x < zoom_y ? zoom_x : zoom_y);
			width = (int)(zoom*m_PhtWidth+0.5);
			height = (int)(zoom*m_PhtHeight+0.5);
			dx = (int)((m_PixWidth - width)/2);
			dy = (int)((m_PixHeight - height)/2);
			// Fill background
			cairo_identity_matrix(cr);
			cairo_set_source_rgb(cr, 0.2, 0.0, 0.4);
			cairo_rectangle(cr, dx, dy, width, height);
			cairo_fill_preserve(cr);
			// Draw stars
			cairo_clip(cr);
			cairo_set_line_width(cr, 1.0);
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_DEFAULT);
			for (i=0; i<m_StCount; i++) {
				const tStarRec *star = &m_Stars[i];
				r = 3.0 + 0.75*(m_RefMag - star->mag);
				r = LimitValue(r, 2.0, 12.0)*sqrt(zoom);
				if (r>1) {
					x = dx + star->x*zoom;
					y = dy + star->y*zoom;
					cairo_arc(cr, x, y, r, 0.0, 2.0*M_PI);
					if (star->type == ST_VARIABLE)
						cairo_set_source_rgb(cr, 0.7, 0, 0);
					else if (star->type == ST_COMPARISON)
						cairo_set_source_rgb(cr, 0, 0.5, 0);
					else if (star->type == ST_CHECK)
						cairo_set_source_rgb(cr, 0, 0, 0.9);
					else
						cairo_set_source_rgb(cr, 1, 1, 1);
					cairo_fill_preserve(cr);
					cairo_set_source_rgb (cr, 0, 0, 0);
					cairo_stroke(cr);
				}
			}
			cairo_reset_clip(cr);
		}

		// Draw table
		if (m_Data) {
			// Canvas
			zoom_x = (double)m_PixWidth/4;
			zoom_y = (double)m_PixHeight/3;
			zoom = (zoom_x < zoom_y ? zoom_x : zoom_y);
			width = (int)(zoom*4+0.5);
			height = (int)(zoom*3+0.5);
			dx = (int)((m_PixWidth - width)/2);
			dy = (int)((m_PixHeight - height)/2);
			// Fill background
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
			cairo_identity_matrix(cr);
			cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
			cairo_rectangle(cr, dx, dy, width, height);
			cairo_fill(cr);
			// Mapping coefficients
			if ((m_MaxX-m_MinX)>DBL_EPSILON && width>10) {
				zoom_x = (width-10)/(m_MaxX-m_MinX);
				offset_x = zoom_x * m_MinX - 5;
			} else {
				zoom_x = 1.0;
				offset_x = (m_MinX+m_MaxX)/2 - (width/2);
			}
			if ((m_MaxY-m_MinY)>DBL_EPSILON && height>10) {
				zoom_y = (height-10)/(m_MinY-m_MaxY);
				offset_y = zoom_y * m_MaxY - 5;
			} else {
				zoom_y = 1.0;
				offset_y = (m_MinY+m_MaxY)/2 - (height/2);
			}
			/* Draw axes
			cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
			cairo_set_line_width(cr, 1.0);
			cairo_move_to(cr, dx+2, dy+height-3);
			cairo_line_to(cr, dx+2, dy+2);
			cairo_move_to(cr, dx+2, dy+height-3);
			cairo_line_to(cr, width-2, dy+height-3);
			cairo_stroke(cr);*/
			// Draw data
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_DEFAULT);
			cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
			cairo_set_line_width(cr, 1.0);
			for (i=0; i<m_DataCount; i++) {
				x = m_Data[i].x * zoom_x - offset_x;
				y = m_Data[i].y * zoom_y - offset_y;
				if (x>=4 && y>=4 && x<=width-4 && y<=height-4) {
					cairo_arc(cr, x+dx, y+dy, 0.5, 0.0, 2.0*M_PI);
					cairo_stroke(cr);
				}
			}
		}

		// Text
		if (m_Text) {
			// Canvas
			zoom_x = (double)m_PixWidth/4;
			zoom_y = (double)m_PixHeight/3;
			zoom = (zoom_x < zoom_y ? zoom_x : zoom_y);
			width = (int)(zoom*4+0.5);
			height = (int)(zoom*3+0.5);
			dx = (int)((m_PixWidth - width)/2);
			dy = (int)((m_PixHeight - height)/2);
			// Fill background
			cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
			cairo_identity_matrix(cr);
			cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
			cairo_rectangle(cr, dx, dy, width, height);
			cairo_fill(cr);
			// Draw text
			cairo_set_font_size(cr, FONT_SIZE);
			cairo_select_font_face(cr, "monospace", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
			cairo_set_source_rgb(cr, 0, 0, 0);
			TextExtents(cr, m_Text, &width, &height);
			DrawText(cr, (m_PixWidth - width)/2, (m_PixHeight - height)/2, 
				width, height, m_Text);
		}
		cairo_destroy(cr);
		gtk_widget_queue_draw(m_DrawArea);
	}
}

gint CPreview::button_press_event(GtkWidget *widget, GdkEventButton *event, CPreview *pMe)
{
	pMe->OnButtonPress(widget, event);
	return true;
}

void CPreview::OnButtonPress(GtkWidget *widget, GdkEventButton *event)
{
	if (event->type==GDK_2BUTTON_PRESS && event->button==1 && m_Path) {
		// Left double click
		ShowFileModal();
	}
}


//
// Open preview in standalone window
//
void CPreview::ShowFileModal(void)
{
	CFileDlgModal *pFileDlg;

	// Autodetect file type
	switch (FileType(m_Path))
	{
	case FILE_TYPE_IMAGE:
		pFileDlg = new CCCDFileDlgModal(m_pParent);
		break;
	case FILE_TYPE_PHOT:
		pFileDlg = new CPhtFileDlgModal(m_pParent);
		break;
	default:
		pFileDlg = NULL;
	}
	if (pFileDlg) {
		// Load file and display it
		pFileDlg->ShowModal(m_Path);
		delete pFileDlg;
	}
}
